# Copyright contributors to the TSFM project
#
"""Basic functions and utilities"""

import copy
import enum
from datetime import datetime
from math import isnan
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from pandas.api.types import is_datetime64_dtype


def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    Source: setuptools/_distutils/util.py
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return 1
    elif val in ("n", "no", "f", "false", "off", "0"):
        return 0
    else:
        raise ValueError(f"invalid truth value {val!r}")


class FractionLocation(enum.Enum):
    """`Enum` for the different methods of choosing the fraction of data. Options first and last are used
    to select the fraction of data from of the begining or end of the full dataset. Uniform chooses windows
    uniformly sampled from the dataset.
    """

    FIRST = "first"
    LAST = "last"
    UNIFORM = "uniform"


def select_by_timestamp(
    df: pd.DataFrame,
    timestamp_column: str = "timestamp",
    start_timestamp: Optional[Union[str, datetime]] = None,
    end_timestamp: Optional[Union[str, datetime]] = None,
) -> pd.DataFrame:
    """Select a portion of a dataset based on timestamps.
    Note that the range selected is inclusive of the starting timestamp.

    Args:
        df (pd.DataFrame): Input dataframe.
        timestamp_column (str, optional): Timestamp column in the dataset. Defaults to "timestamp".
        start_timestamp (Optional[Union[str, datetime]], optional): Timestamp of the starting point.
            Defaults to None. Use None to specify the start of the data.
        end_timestamp (Optional[Union[str, datetime]], optional): Timestamp of the ending point.
            Use None to specify the end of the data. Defaults to None.

    Raises:
        ValueError: User must specify either start_timestamp or end_timestamp.

    Returns:
        pd.DataFrame: Subset of the dataframe.
    """

    if not start_timestamp and not end_timestamp:
        raise ValueError("At least one of start_timestamp or end_timestamp must be specified.")

    if not start_timestamp:
        return df[df[timestamp_column] < end_timestamp]

    if not end_timestamp:
        return df[df[timestamp_column] >= start_timestamp]

    return df[(df[timestamp_column] >= start_timestamp) & (df[timestamp_column] < end_timestamp)]


def select_by_index(
    df: pd.DataFrame,
    id_columns: Optional[List[str]] = None,
    start_index: Optional[int] = None,
    end_index: Optional[int] = None,
) -> pd.DataFrame:
    """Select a portion of a dataset based on integer indices into the data.
    Note that the range selected is inclusive of the starting index, but exclusive of the end index. When ID
    columns are specified the selection is done per-time series (i.e., the indices are used relative to each
    time series). The indexing is intended to be similar to python-style indexing into lists, where the end
    of the specified range is not included.

    Args:
        df (pd.DataFrame): Input dataframe.
        id_columns (List[str], optional): Columns which specify the IDs in the dataset. Defaults to None.
        start_index (Optional[int], optional): Index of the starting point.
            Defaults to None. Use None to specify the start of the data.
        end_index (Optional[Union[str, datetime]], optional): Index for the end of the selection (not inclusive).
            Use None to specify the end of the data. Defaults to None.

    Raises:
        ValueError: User must specify either start_index or end_index.

    Returns:
        pd.DataFrame: Subset of the dataframe.
    """
    if not start_index and not end_index:
        raise ValueError("At least one of start_index or end_index must be specified.")

    if not id_columns:
        return _split_group_by_index(df, start_index=start_index, end_index=end_index).copy()

    groups = df.groupby(_get_groupby_columns(id_columns))
    result = []
    for name, group in groups:
        result.append(_split_group_by_index(group, name=name, start_index=start_index, end_index=end_index))

    return pd.concat(result)


def select_by_relative_fraction(
    df: pd.DataFrame,
    id_columns: Optional[List[str]] = None,
    start_fraction: Optional[float] = None,
    start_offset: Optional[int] = 0,
    end_fraction: Optional[float] = None,
) -> pd.DataFrame:
    """Select a portion of a dataset based on relative fractions of the data.
    Note that the range selected is inclusive of the starting index. When ID columns are specified
    the selection is done per-time series (i.e., the fractions are used relative to each time series length).

    The indices are computed as:
    index_start_i = floor(length_i * start_fraction) - start_offset
    index_end_i = floor(length_i * end_fraction)

    Args:
        df (pd.DataFrame): Input dataframe.
        id_columns (List[str], optional): Columns which specify the IDs in the dataset. Defaults to None.
        start_fraction (Optional[float], optional): The fraction to specify the start of the selection. Use None to specify the start of the dataset. Defaults to None.
        start_offset (Optional[int], optional): An optional offset to apply to the starting point of
            each subseries. A non-negative value should be used. Defaults to 0.
        end_fraction (Optional[float], optional): The fraction to specify the end of the selection.
            Use None to specify the end of the dataset. Defaults to None.

    Raises:
        ValueError: Raised when the user does not specify either start_index or end_index. Also raised
            when a negative value of start_offset is provided.

    Returns:
        pd.DataFrame: Subset of the dataframe.
    """
    if not start_fraction and not end_fraction:
        raise ValueError("At least one of start_fraction or end_fraction must be specified.")

    if start_offset < 0:
        raise ValueError("The value of start_offset should ne non-negative.")

    if not id_columns:
        return _split_group_by_fraction(
            df,
            start_fraction=start_fraction,
            end_fraction=end_fraction,
            start_offset=start_offset,
        ).copy()

    groups = df.groupby(_get_groupby_columns(id_columns))
    result = []
    for name, group in groups:
        result.append(
            _split_group_by_fraction(
                group,
                name=name,
                start_fraction=start_fraction,
                end_fraction=end_fraction,
                start_offset=start_offset,
            )
        )

    return pd.concat(result)


def select_by_fixed_fraction(
    df: pd.DataFrame,
    id_columns: Optional[List[str]] = None,
    fraction: float = 1.0,
    location: str = FractionLocation.FIRST.value,
    minimum_size: Optional[int] = 0,
) -> pd.DataFrame:
    """Select a portion of a dataset based on a fraction of the data.
    Fraction can either be located at the start (location = FractionLocation.FIRST) or at the end (location = FractionLocation.LAST)

    Args:
        df (pd.DataFrame): Input dataframe.
        id_columns (List[str], optional): Columns which specify the IDs in the dataset. Defaults to None.
        fraction (float): The fraction to select.
        location (str): Location of where to select the fraction Defaults to FractionLocation.FIRST.value.
        minimum_size (int, optional): Minimum size of the split. Defaults to None.

    Raises:
        ValueError: Raised when the fraction is not within the range [0,1].

    Returns:
        pd.DataFrame: Subset of the dataframe.
    """

    if fraction < 0 or fraction > 1:
        raise ValueError("The value of fraction should be between 0 and 1.")

    if not id_columns:
        return _split_group_by_fixed_fraction(
            df, fraction=fraction, location=location, minimum_size=minimum_size
        ).copy()

    groups = df.groupby(_get_groupby_columns(id_columns))
    result = []
    for name, group in groups:
        result.append(
            _split_group_by_fixed_fraction(
                group,
                name=name,
                fraction=fraction,
                location=location,
                minimum_size=minimum_size,
            )
        )

    return pd.concat(result)


def train_test_split(
    df: pd.DataFrame,
    id_columns: Optional[List[str]] = None,
    train: Union[int, float] = 0.7,
    test: Union[int, float] = 0.2,
    valid_test_offset: int = 0,
):
    # to do: add validation

    if not id_columns:
        return tuple(
            [
                tmp.copy()
                for tmp in _split_group_train_test(df, train=train, test=test, valid_test_offset=valid_test_offset)
            ]
        )

    groups = df.groupby(_get_groupby_columns(id_columns))
    result = []
    for name, group in groups:
        result.append(
            _split_group_train_test(
                group,
                name=name,
                train=train,
                test=test,
                valid_test_offset=valid_test_offset,
            )
        )

    result_train, result_valid, result_test = zip(*result)
    return pd.concat(result_train), pd.concat(result_valid), pd.concat(result_test)


def _split_group_train_test(
    group_df: pd.DataFrame,
    name: Optional[str] = None,
    train: Union[int, float] = 0.7,
    test: Union[int, float] = 0.2,
    valid_test_offset: int = 0,
):
    l = len(group_df)

    train_size = int(l * train)
    test_size = int(l * test)

    valid_size = l - train_size - test_size

    train_df = _split_group_by_index(group_df, name, start_index=0, end_index=train_size)

    valid_df = _split_group_by_index(
        group_df,
        name,
        start_index=train_size - valid_test_offset,
        end_index=train_size + valid_size,
    )

    test_df = _split_group_by_index(group_df, name, start_index=train_size + valid_size - valid_test_offset)

    return train_df, valid_df, test_df


def _get_groupby_columns(id_columns: List[str]) -> Union[List[str], str]:
    if not isinstance(id_columns, (List)):
        raise ValueError("id_columns must be a list")

    if len(id_columns) == 1:
        return id_columns[0]

    return id_columns


def _split_group_by_index(
    group_df: pd.DataFrame,
    name: Optional[str] = None,
    start_index: Optional[int] = None,
    end_index: Optional[int] = None,
) -> pd.DataFrame:
    """Helper function for splitting by index."""
    if start_index and (start_index >= len(group_df)):
        msg = "Selection would result in an empty time series, please check start_index and time series length"
        msg = msg + f" (id = {name})" if name else msg
        raise ValueError(msg)

    # Also check that end_index <= len(group_df)?

    if not start_index:
        return group_df.iloc[:end_index,]

    if not end_index:
        return group_df.iloc[start_index:,]

    return group_df.iloc[start_index:end_index, :]


def _split_group_by_fraction(
    group_df: pd.DataFrame,
    name: Optional[str] = None,
    start_fraction: Optional[float] = None,
    start_offset: Optional[int] = 0,
    end_fraction: Optional[float] = None,
) -> pd.DataFrame:
    """Helper function for splitting by relative fraction."""
    length = len(group_df)

    if start_fraction is not None:
        start_index = int(length * start_fraction) - start_offset
        if start_index < 0:
            if name:
                msg = f"Computed starting_index for id={name} is negative, please check individual time series lengths, start_fraction, and start_offset."
            else:
                msg = "Computed starting_index is negative, please check time series length, start_fraction, and start_offset."
            raise ValueError(msg)
    else:
        start_index = None

    if end_fraction is not None:
        end_index = int(length * end_fraction)
    else:
        end_index = None

    return _split_group_by_index(group_df=group_df, name=name, start_index=start_index, end_index=end_index)


def _split_group_by_fixed_fraction(
    group_df: pd.DataFrame,
    name: Optional[str] = None,
    fraction: float = 1.0,
    location: Optional[str] = None,
    minimum_size: Optional[int] = 0,
):
    """Helper function for splitting by fixed fraction."""
    l = len(group_df)
    fraction_size = int(fraction * (l - minimum_size)) + minimum_size

    if location == FractionLocation.FIRST.value:
        start_index = 0
        end_index = fraction_size
    elif location == FractionLocation.LAST.value:
        start_index = l - fraction_size
        end_index = l
    else:
        raise ValueError(
            f"`location` should be either `{FractionLocation.FIRST.value}` or `{FractionLocation.LAST.value}`"
        )

    return _split_group_by_index(group_df=group_df, name=name, start_index=start_index, end_index=end_index)


def convert_tsf_to_dataframe(
    full_file_path_and_name,
    replace_missing_vals_with="NaN",
    value_column_name="series_value",
):
    """Read a .tsf file into a pandas dataframe.

    Args:
        full_file_path_and_name (_type_): _description_
        replace_missing_vals_with (str, optional): _description_. Defaults to "NaN".
        value_column_name (str, optional): _description_. Defaults to "series_value".

    This code adopted from the Monash forecasting repository github:
    https://github.com/rakshitha123/TSForecasting/blob/master/utils/data_loader.py

    """

    col_names = []
    col_types = []
    all_data = {}
    line_count = 0
    frequency = None
    forecast_horizon = None
    contain_missing_values = None
    contain_equal_length = None
    found_data_tag = False
    found_data_section = False
    started_reading_data_section = False

    with open(full_file_path_and_name, "r", encoding="cp1252") as file:
        for line in file:
            # Strip white space from start/end of line
            line = line.strip()

            if line:
                if line.startswith("@"):  # Read meta-data
                    if not line.startswith("@data"):
                        line_content = line.split(" ")
                        if line.startswith("@attribute"):
                            if len(line_content) != 3:  # Attributes have both name and type
                                raise Exception("Invalid meta-data specification.")

                            col_names.append(line_content[1])
                            col_types.append(line_content[2])
                        else:
                            if len(line_content) != 2:  # Other meta-data have only values
                                raise Exception("Invalid meta-data specification.")

                            if line.startswith("@frequency"):
                                frequency = line_content[1]
                            elif line.startswith("@horizon"):
                                forecast_horizon = int(line_content[1])
                            elif line.startswith("@missing"):
                                contain_missing_values = bool(strtobool(line_content[1]))
                            elif line.startswith("@equallength"):
                                contain_equal_length = bool(strtobool(line_content[1]))

                    else:
                        if len(col_names) == 0:
                            raise Exception("Missing attribute section. Attribute section must come before data.")

                        found_data_tag = True
                elif not line.startswith("#"):
                    if len(col_names) == 0:
                        raise Exception("Missing attribute section. Attribute section must come before data.")
                    elif not found_data_tag:
                        raise Exception("Missing @data tag.")
                    else:
                        if not started_reading_data_section:
                            started_reading_data_section = True
                            found_data_section = True
                            all_series = []

                            for col in col_names:
                                all_data[col] = []

                        full_info = line.split(":")

                        if len(full_info) != (len(col_names) + 1):
                            raise Exception("Missing attributes/values in series.")

                        series = full_info[len(full_info) - 1]
                        series = series.split(",")

                        if len(series) == 0:
                            raise Exception(
                                "A given series should contains a set of comma separated numeric values. At least one numeric value should be there in a series. Missing values should be indicated with ? symbol"
                            )

                        numeric_series = []

                        for val in series:
                            if val == "?":
                                numeric_series.append(replace_missing_vals_with)
                            else:
                                numeric_series.append(float(val))

                        if numeric_series.count(replace_missing_vals_with) == len(numeric_series):
                            raise Exception(
                                "All series values are missing. A given series should contains a set of comma separated numeric values. At least one numeric value should be there in a series."
                            )

                        all_series.append(pd.Series(numeric_series).array)

                        for i in range(len(col_names)):
                            att_val = None
                            if col_types[i] == "numeric":
                                att_val = int(full_info[i])
                            elif col_types[i] == "string":
                                att_val = str(full_info[i])
                            elif col_types[i] == "date":
                                att_val = datetime.strptime(full_info[i], "%Y-%m-%d %H-%M-%S")
                            else:
                                raise Exception(
                                    "Invalid attribute type."
                                )  # Currently, the code supports only numeric, string and date types. Extend this as required.

                            if att_val is None:
                                raise Exception("Invalid attribute value.")
                            else:
                                all_data[col_names[i]].append(att_val)

                line_count = line_count + 1

        if line_count == 0:
            raise Exception("Empty file.")
        if len(col_names) == 0:
            raise Exception("Missing attribute section.")
        if not found_data_section:
            raise Exception("Missing series information under data section.")

        all_data[value_column_name] = all_series
        loaded_data = pd.DataFrame(all_data)

        return (
            loaded_data,
            frequency,
            forecast_horizon,
            contain_missing_values,
            contain_equal_length,
        )


def convert_tsfile_to_dataframe(
    full_file_path_and_name,
    return_separate_X_and_y=True,
    replace_missing_vals_with="NaN",
):
    """Load data from a .ts file into a Pandas DataFrame.
    Parameters
    ----------
    full_file_path_and_name: str
        The full pathname of the .ts file to read.
    return_separate_X_and_y: bool
        true if X and Y values should be returned as separate Data Frames (
        X) and a numpy array (y), false otherwise.
        This is only relevant for data that
    replace_missing_vals_with: str
       The value that missing values in the text file should be replaced
       with prior to parsing.
    Returns
    -------
    DataFrame (default) or ndarray (i
        If return_separate_X_and_y then a tuple containing a DataFrame and a
        numpy array containing the relevant time-series and corresponding
        class values.
    DataFrame
        If not return_separate_X_and_y then a single DataFrame containing
        all time-series and (if relevant) a column "class_vals" the
        associated class values.


    This code adopted from sktime:
    https://github.com/sktime/sktime/blob/v0.30.0/sktime/datasets/_readers_writers/ts.py#L32-L615

    """
    # Initialize flags and variables used when parsing the file
    metadata_started = False
    data_started = False

    has_problem_name_tag = False
    has_timestamps_tag = False
    has_univariate_tag = False
    has_class_labels_tag = False
    has_data_tag = False

    previous_timestamp_was_int = None
    prev_timestamp_was_timestamp = None
    num_dimensions = None
    is_first_case = True
    instance_list = []
    class_val_list = []
    line_num = 0
    # Parse the file
    with open(full_file_path_and_name, "r", encoding="utf-8") as file:
        for line in file:
            # Strip white space from start/end of line and change to
            # lowercase for use below
            line = line.strip().lower()
            # Empty lines are valid at any point in a file
            if line:
                # Check if this line contains metadata
                # Please note that even though metadata is stored in this
                # function it is not currently published externally
                if line.startswith("@problemname"):
                    # Check that the data has not started
                    if data_started:
                        raise IOError("metadata must come before data")
                    # Check that the associated value is valid
                    tokens = line.split(" ")
                    token_len = len(tokens)
                    if token_len == 1:
                        raise IOError("problemname tag requires an associated value")
                    # problem_name = line[len("@problemname") + 1:]
                    has_problem_name_tag = True
                    metadata_started = True
                elif line.startswith("@timestamps"):
                    # Check that the data has not started
                    if data_started:
                        raise IOError("metadata must come before data")
                    # Check that the associated value is valid
                    tokens = line.split(" ")
                    token_len = len(tokens)
                    if token_len != 2:
                        raise IOError("timestamps tag requires an associated Boolean value")
                    elif tokens[1] == "true":
                        timestamps = True
                    elif tokens[1] == "false":
                        timestamps = False
                    else:
                        raise IOError("invalid timestamps value")
                    has_timestamps_tag = True
                    metadata_started = True
                elif line.startswith("@univariate"):
                    # Check that the data has not started
                    if data_started:
                        raise IOError("metadata must come before data")
                    # Check that the associated value is valid
                    tokens = line.split(" ")
                    token_len = len(tokens)
                    if token_len != 2:
                        raise IOError("univariate tag requires an associated Boolean  value")
                    elif tokens[1] == "true":
                        # univariate = True
                        pass
                    elif tokens[1] == "false":
                        # univariate = False
                        pass
                    else:
                        raise IOError("invalid univariate value")
                    has_univariate_tag = True
                    metadata_started = True
                elif line.startswith("@classlabel"):
                    # Check that the data has not started
                    if data_started:
                        raise IOError("metadata must come before data")
                    # Check that the associated value is valid
                    tokens = line.split(" ")
                    token_len = len(tokens)
                    if token_len == 1:
                        raise IOError("classlabel tag requires an associated Boolean  value")
                    if tokens[1] == "true":
                        class_labels = True
                    elif tokens[1] == "false":
                        class_labels = False
                    else:
                        raise IOError("invalid classLabel value")
                    # Check if we have any associated class values
                    if token_len == 2 and class_labels:
                        raise IOError("if the classlabel tag is true then class values must be supplied")
                    has_class_labels_tag = True
                    class_label_list = [token.strip() for token in tokens[2:]]
                    metadata_started = True
                    classification_dataset = True
                elif line.startswith("@targetlabel"):
                    if data_started:
                        raise IOError("metadata must come before data")
                    tokens = line.split(" ")
                    token_len = len(tokens)
                    if token_len == 1:
                        raise IOError("targetlabel tag requires an associated Boolean value")
                    if tokens[1] == "true":
                        class_labels = True
                    elif tokens[1] == "false":
                        class_labels = False
                    else:
                        raise IOError("invalid targetlabel value")
                    if token_len > 2:
                        raise IOError(
                            "targetlabel tag should not be accompanied with info "
                            "apart from true/false, but found "
                            f"{tokens}"
                        )
                    has_class_labels_tag = True
                    metadata_started = True
                    classification_dataset = False
                # Check if this line contains the start of data
                elif line.startswith("@data"):
                    if line != "@data":
                        raise IOError("data tag should not have an associated value")
                    if data_started and not metadata_started:
                        raise IOError("metadata must come before data")
                    else:
                        has_data_tag = True
                        data_started = True
                # If the 'data tag has been found then metadata has been
                # parsed and data can be loaded
                elif data_started:
                    # Check that a full set of metadata has been provided
                    if (
                        not has_problem_name_tag
                        or not has_timestamps_tag
                        or not has_univariate_tag
                        or not has_class_labels_tag
                        or not has_data_tag
                    ):
                        raise IOError("a full set of metadata has not been provided before the data")
                    # Replace any missing values with the value specified
                    line = line.replace("?", replace_missing_vals_with)
                    # Check if we are dealing with data that has timestamps
                    if timestamps:
                        # We're dealing with timestamps so cannot just split
                        # line on ':' as timestamps may contain one
                        has_another_value = False
                        has_another_dimension = False
                        timestamp_for_dim = []
                        values_for_dimension = []
                        this_line_num_dim = 0
                        line_len = len(line)
                        char_num = 0
                        while char_num < line_len:
                            # Move through any spaces
                            while char_num < line_len and str.isspace(line[char_num]):
                                char_num += 1
                            # See if there is any more data to read in or if
                            # we should validate that read thus far
                            if char_num < line_len:
                                # See if we have an empty dimension (i.e. no
                                # values)
                                if line[char_num] == ":":
                                    if len(instance_list) < (this_line_num_dim + 1):
                                        instance_list.append([])
                                    instance_list[this_line_num_dim].append(pd.Series(dtype="object"))
                                    this_line_num_dim += 1
                                    has_another_value = False
                                    has_another_dimension = True
                                    timestamp_for_dim = []
                                    values_for_dimension = []
                                    char_num += 1
                                else:
                                    # Check if we have reached a class label
                                    if line[char_num] != "(" and class_labels:
                                        class_val = line[char_num:].strip()
                                        if classification_dataset:
                                            if class_val not in class_label_list:
                                                raise IOError(
                                                    "the class value '"
                                                    + class_val
                                                    + "' on line "
                                                    + str(line_num + 1)
                                                    + " is not "
                                                    "valid"
                                                )
                                        class_val_list.append(class_val)
                                        char_num = line_len
                                        has_another_value = False
                                        has_another_dimension = False
                                        timestamp_for_dim = []
                                        values_for_dimension = []
                                    else:
                                        # Read in the data contained within
                                        # the next tuple
                                        if line[char_num] != "(" and not class_labels:
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " does "
                                                "not "
                                                "start "
                                                "with a "
                                                "'('"
                                            )
                                        char_num += 1
                                        tuple_data = ""
                                        while char_num < line_len and line[char_num] != ")":
                                            tuple_data += line[char_num]
                                            char_num += 1
                                        if char_num >= line_len or line[char_num] != ")":
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " does "
                                                "not end"
                                                " with a "
                                                "')'"
                                            )
                                        # Read in any spaces immediately
                                        # after the current tuple
                                        char_num += 1
                                        while char_num < line_len and str.isspace(line[char_num]):
                                            char_num += 1

                                        # Check if there is another value or
                                        # dimension to process after this tuple
                                        if char_num >= line_len:
                                            has_another_value = False
                                            has_another_dimension = False
                                        elif line[char_num] == ",":
                                            has_another_value = True
                                            has_another_dimension = False
                                        elif line[char_num] == ":":
                                            has_another_value = False
                                            has_another_dimension = True
                                        char_num += 1
                                        # Get the numeric value for the
                                        # tuple by reading from the end of
                                        # the tuple data backwards to the
                                        # last comma
                                        last_comma_index = tuple_data.rfind(",")
                                        if last_comma_index == -1:
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " contains a tuple that has "
                                                "no comma inside of it"
                                            )
                                        try:
                                            value = tuple_data[last_comma_index + 1 :]
                                            value = float(value)
                                        except ValueError:
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " contains a tuple that does "
                                                "not have a valid numeric "
                                                "value"
                                            )
                                        # Check the type of timestamp that
                                        # we have
                                        timestamp = tuple_data[0:last_comma_index]
                                        try:
                                            timestamp = int(timestamp)
                                            timestamp_is_int = True
                                            timestamp_is_timestamp = False
                                        except ValueError:
                                            timestamp_is_int = False
                                        if not timestamp_is_int:
                                            try:
                                                timestamp = timestamp.strip()
                                                timestamp_is_timestamp = True
                                            except ValueError:
                                                timestamp_is_timestamp = False
                                        # Make sure that the timestamps in
                                        # the file (not just this dimension
                                        # or case) are consistent
                                        if not timestamp_is_timestamp and not timestamp_is_int:
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " contains a tuple that "
                                                "has an invalid timestamp '" + timestamp + "'"
                                            )
                                        if (
                                            previous_timestamp_was_int is not None
                                            and previous_timestamp_was_int
                                            and not timestamp_is_int
                                        ):
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " contains tuples where the "
                                                "timestamp format is "
                                                "inconsistent"
                                            )
                                        if (
                                            prev_timestamp_was_timestamp is not None
                                            and prev_timestamp_was_timestamp
                                            and not timestamp_is_timestamp
                                        ):
                                            raise IOError(
                                                "dimension "
                                                + str(this_line_num_dim + 1)
                                                + " on line "
                                                + str(line_num + 1)
                                                + " contains tuples where the "
                                                "timestamp format is "
                                                "inconsistent"
                                            )
                                        # Store the values
                                        timestamp_for_dim += [timestamp]
                                        values_for_dimension += [value]
                                        #  If this was our first tuple then
                                        #  we store the type of timestamp we
                                        #  had
                                        if prev_timestamp_was_timestamp is None and timestamp_is_timestamp:
                                            prev_timestamp_was_timestamp = True
                                            previous_timestamp_was_int = False

                                        if previous_timestamp_was_int is None and timestamp_is_int:
                                            prev_timestamp_was_timestamp = False
                                            previous_timestamp_was_int = True
                                        # See if we should add the data for
                                        # this dimension
                                        if not has_another_value:
                                            if len(instance_list) < (this_line_num_dim + 1):
                                                instance_list.append([])

                                            if timestamp_is_timestamp:
                                                timestamp_for_dim = pd.DatetimeIndex(timestamp_for_dim)

                                            instance_list[this_line_num_dim].append(
                                                pd.Series(
                                                    index=timestamp_for_dim,
                                                    data=values_for_dimension,
                                                )
                                            )
                                            this_line_num_dim += 1
                                            timestamp_for_dim = []
                                            values_for_dimension = []
                            elif has_another_value:
                                raise IOError(
                                    "dimension " + str(this_line_num_dim + 1) + " on "
                                    "line " + str(line_num + 1) + " ends with a ',' that "
                                    "is not followed by "
                                    "another tuple"
                                )
                            elif has_another_dimension and class_labels:
                                raise IOError(
                                    "dimension " + str(this_line_num_dim + 1) + " on "
                                    "line " + str(line_num + 1) + " ends with a ':' while "
                                    "it should list a class "
                                    "value"
                                )
                            elif has_another_dimension and not class_labels:
                                if len(instance_list) < (this_line_num_dim + 1):
                                    instance_list.append([])
                                instance_list[this_line_num_dim].append(pd.Series(dtype=np.float32))
                                this_line_num_dim += 1
                                num_dimensions = this_line_num_dim
                            # If this is the 1st line of data we have seen
                            # then note the dimensions
                            if not has_another_value and not has_another_dimension:
                                if num_dimensions is None:
                                    num_dimensions = this_line_num_dim
                                if num_dimensions != this_line_num_dim:
                                    raise IOError(
                                        "line " + str(line_num + 1) + " does not have the "
                                        "same number of "
                                        "dimensions as the "
                                        "previous line of "
                                        "data"
                                    )
                        # Check that we are not expecting some more data,
                        # and if not, store that processed above
                        if has_another_value:
                            raise IOError(
                                "dimension "
                                + str(this_line_num_dim + 1)
                                + " on line "
                                + str(line_num + 1)
                                + " ends with a ',' that is "
                                "not followed by another "
                                "tuple"
                            )
                        elif has_another_dimension and class_labels:
                            raise IOError(
                                "dimension "
                                + str(this_line_num_dim + 1)
                                + " on line "
                                + str(line_num + 1)
                                + " ends with a ':' while it "
                                "should list a class value"
                            )
                        elif has_another_dimension and not class_labels:
                            if len(instance_list) < (this_line_num_dim + 1):
                                instance_list.append([])
                            instance_list[this_line_num_dim].append(pd.Series(dtype="object"))
                            this_line_num_dim += 1
                            num_dimensions = this_line_num_dim
                        # If this is the 1st line of data we have seen then
                        # note the dimensions
                        if not has_another_value and num_dimensions != this_line_num_dim:
                            raise IOError(
                                "line " + str(line_num + 1) + " does not have the same "
                                "number of dimensions as the "
                                "previous line of data"
                            )
                        # Check if we should have class values, and if so
                        # that they are contained in those listed in the
                        # metadata
                        if class_labels and len(class_val_list) == 0:
                            raise IOError("the cases have no associated class values")
                    else:
                        dimensions = line.split(":")
                        # If first row then note the number of dimensions (
                        # that must be the same for all cases)
                        if is_first_case:
                            num_dimensions = len(dimensions)
                            if class_labels:
                                num_dimensions -= 1
                            for _dim in range(0, num_dimensions):
                                instance_list.append([])
                            is_first_case = False
                        # See how many dimensions that the case whose data
                        # in represented in this line has
                        this_line_num_dim = len(dimensions)
                        if class_labels:
                            this_line_num_dim -= 1
                        # All dimensions should be included for all series,
                        # even if they are empty
                        if this_line_num_dim != num_dimensions:
                            raise IOError(
                                "inconsistent number of dimensions. "
                                "Expecting " + str(num_dimensions) + " but have read " + str(this_line_num_dim)
                            )
                        # Process the data for each dimension
                        for dim in range(0, num_dimensions):
                            dimension = dimensions[dim].strip()

                            if dimension:
                                data_series = dimension.split(",")
                                data_series = [float(i) for i in data_series]
                                instance_list[dim].append(pd.Series(data_series))
                            else:
                                instance_list[dim].append(pd.Series(dtype="object"))
                        if class_labels:
                            class_val_list.append(dimensions[num_dimensions].strip())
            line_num += 1
    # Check that the file was not empty
    if line_num:
        # Check that the file contained both metadata and data
        if metadata_started and not (
            has_problem_name_tag
            and has_timestamps_tag
            and has_univariate_tag
            and has_class_labels_tag
            and has_data_tag
        ):
            raise IOError("metadata incomplete")

        elif metadata_started and not data_started:
            raise IOError("file contained metadata but no data")

        elif metadata_started and data_started and len(instance_list) == 0:
            raise IOError("file contained metadata but no data")
        # Create a DataFrame from the data parsed above
        data = pd.DataFrame(dtype=np.float32)
        for dim in range(0, num_dimensions):
            data["dim_" + str(dim)] = instance_list[dim]
        # Check if we should return any associated class labels separately
        if class_labels:
            if return_separate_X_and_y:
                return data, np.asarray(class_val_list)
            else:
                data["class_vals"] = pd.Series(class_val_list)
                return data
        else:
            return data
    else:
        raise IOError("empty file")


def get_split_params(
    split_config: Dict[str, Union[float, List[Union[int, float]]]],
    context_length: Optional[int] = None,
) -> Tuple[Dict[str, Dict[str, Union[int, float]]], Dict[str, Callable]]:
    """Get split parameters

    Args:
        split_config ( Dict[str, Union[float, List[Union[int, float]]]]): Dictionary containing keys which
            define the splits. Two options are possible:
            1. Specifiy train, valid, test. Each value consists of a list of length two, indicating
            the boundaries of a split.
            2. Specify train, test. Each value consists of a single floating point number specifying the
            fraction of data to use. Valid is populated using the remaining data.

        context_length (int, optional): Context length, used only when offseting
            the split so predictions can be made for all elements of split. Defaults to None.

    Returns:
        Tuple[Dict[str, Dict[str, Union[int, float]]], Dict[str, Callable]]: Tuple of split parameters
        and split functions to use to split the data.
    """

    split_params = {}
    split_function = {}

    if "valid" in split_config:
        for group in ["train", "test", "valid"]:
            if ((split_config[group][0] < 1) and (split_config[group][0] != 0)) or (split_config[group][1] < 1):
                split_params[group] = {
                    "start_fraction": split_config[group][0],
                    "end_fraction": split_config[group][1],
                    "start_offset": (context_length if (context_length and group != "train") else 0),
                }
                split_function[group] = select_by_relative_fraction
            else:
                split_params[group] = {
                    "start_index": (
                        split_config[group][0] - (context_length if (context_length and group != "train") else 0)
                    ),
                    "end_index": split_config[group][1],
                }
                split_function[group] = select_by_index
        return split_params, split_function

    # no valid, assume train/test split
    split_function = train_test_split
    split_params = {
        "train": split_config["train"],
        "test": split_config["test"],
        "valid_test_offset": context_length if context_length else 0,
    }
    return split_params, split_function


def convert_tsf(filename: str) -> pd.DataFrame:
    """Converts a tsf format file into a pandas dataframe.
    Returns the result in canonical multi-time series format, with an ID column, timestamp, and one or more
    value columns. Attemps to map frequency information given in the input file to pandas equivalents.


    Args:
        filename (str): Input file name.

    Returns:
        pd.DataFrame: Converted time series
    """
    (
        loaded_data,
        frequency,
        forecast_horizon,
        contain_missing_values,
        contain_equal_length,
    ) = convert_tsf_to_dataframe(filename, replace_missing_vals_with=np.nan)

    id_column_name = "id"
    timestamp_column_name = "timestamp"
    value_column_name = "value"

    tsf_to_pandas_freq_map = {
        "daily": "d",
        "hourly": "h",
        "half_hourly": "30min",
        "seconds": "s",
        "minutes": "min",
        "minutely": "min",
        "weekly": "W",
        "monthly": "MS",
        "yearly": "YS",
        "quarterly": "QS",
    }

    if frequency:
        try:
            freq = tsf_to_pandas_freq_map[frequency]
        except KeyError:
            try:
                freq_val, freq_unit = frequency.split("_")
                freq = freq_val + tsf_to_pandas_freq_map[freq_unit]
            except ValueError:
                freq = tsf_to_pandas_freq_map[frequency]
            except KeyError:
                raise ValueError(f"Input file contains an unknown frequency unit {freq_unit}")
    else:
        freq = None

    # determine presence of timestamp column and name
    default_start_timestamp = datetime(1900, 1, 1)
    datetimes = [is_datetime64_dtype(d) for d in loaded_data.dtypes]
    source_timestamp_column = None
    if any(datetimes):
        source_timestamp_column = loaded_data.columns[datetimes][0]

    dfs = []
    for index, item in loaded_data.iterrows():
        if freq and source_timestamp_column:
            timestamps = pd.date_range(item[source_timestamp_column], periods=len(item.series_value), freq=freq)
        elif freq:
            timestamps = pd.date_range(default_start_timestamp, periods=len(item.series_value), freq=freq)
        else:
            timestamps = range(len(item.series_value))

        dfs.append(
            pd.DataFrame(
                {
                    id_column_name: item.series_name,
                    timestamp_column_name: timestamps,
                    value_column_name: item.series_value,
                }
            )
        )

    df = pd.concat(dfs)
    df.reset_index(inplace=True, drop=True)
    return df


def convert_to_univariate(
    data: pd.DataFrame,
    timestamp_column: str,
    id_columns: List[str],
    target_columns: List[str],
    var_name: str = "column_id",
    value_name: str = "value",
) -> pd.DataFrame:
    """Converts a dataframe in canonical format to a univariate dataset. Adds an additional id column to
    indicate the original target column to which a given value corresponds.

    Args:
        data (pd.DataFrame): Input data frame containing multiple target columns.
        timestamp_column (str): String representing the timestamp column.
        id_columns (List[str]): List of columns representing the ids in the data. Use empty list (`[]`) if there
            are no id columns.
        target_columns (List[str]): The target columns in the data.
        var_name (str): Name of new id column used to identify original column name. Defaults to "column_id".
        value_name (str): Name of new value column in the resulting univariate datset. Defaults to "value".

    Returns:
        pd.DataFrame: Converted dataframe.
    """

    if len(target_columns) < 2:
        raise ValueError("`target_columns` should be a non-empty list of two or more elements.")

    return pd.melt(
        data,
        id_vars=[
            timestamp_column,
        ]
        + id_columns,
        value_vars=target_columns,
        var_name=var_name,
        value_name=value_name,
    )


def join_list_without_repeat(*lists: List[List[Any]]) -> List[Any]:
    """Join multiple lists in sequence without repeating

    Returns:
        List[Any]: Combined list.
    """

    final = None
    final_set = set()
    for alist in lists:
        if final is None:
            final = copy.copy(alist)
        else:
            final = final + [item for item in alist if item not in final_set]
        final_set = set(final)
    return final


def count_parameters(model: torch.nn.Module) -> int:
    """Count trainable parameters in a model

    Args:
        model (torch.nn.Module): The model.

    Returns:
        int: Number of parameters requiring gradients.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def convert_tsfile(filename: str) -> pd.DataFrame:
    """Converts a .ts file into a pandas dataframe.
    Returns the result in canonical multi-time series format, with an ID column, and timestamp.

    Args:
        filename (str): Input file name.

    Returns:
        pd.DataFrame: Converted time series


    To do:
    - address renaming of columns
    - check that we catch all timestamp column types

    """

    dfs = []
    df = convert_tsfile_to_dataframe(filename, return_separate_X_and_y=False)

    # rows, columns = df.shape
    value_columns = [c for c in df.columns if c != "class_vals"]

    for row in df.itertuples():
        l = len(row.dim_0)
        temp_df = pd.DataFrame({"id": [row.Index] * l})

        for j, c in enumerate(value_columns):
            c_data = getattr(row, c)
            if isinstance(c_data.index, pd.Timestamp) and "timestamp" not in temp_df.columns:
                ## include timestamp columns if data includes timestamps
                temp_df["timestamp"] = c_data.index
            temp_df[f"value_{j}"] = c_data.values

        temp_df["target"] = row.class_vals

        dfs.append(temp_df)

    final_df = pd.concat(dfs, ignore_index=True)

    # to be moved to a preprocessor
    # ## convert targets to floats or integers
    # ## non-numeric classification labels will be converted to integers as well
    # try:
    #     final_df["target"] = pd.to_numeric(final_df["target"])
    # except KeyError:
    #     string_labels = final_df["target"].unique()
    #     label_to_int_map = {str_label: num for num, str_label in enumerate(string_labels)}
    #     final_df["target"] = final_df["target"].map(label_to_int_map)

    # ## make sure labels are 0 indexed if classification
    # if classification and final_df["target"].min() != 0:
    #     final_df["target"] = final_df["target"] - 1

    return final_df


def encode_data(df: pd.DataFrame, timestamp_column: str) -> Dict[str, Any]:
    """Encode data for sending when using services

    Args:
        df (pd.DataFrame): Input time series data
        timestamp_column (str): Column in the input pandas dataframe where timestamp information resides.

    Returns:
        Dict[str, Any]: _description_
    """
    if pd.api.types.is_datetime64_dtype(df[timestamp_column]):
        df[timestamp_column] = df[timestamp_column].apply(lambda x: x.isoformat())
    data_payload = df.to_dict(orient="list")

    for k, v in data_payload.items():
        if isinstance(v[0], str):
            continue
        # rewrite all the nan to None
        data_payload[k] = [vv if (vv is None) or (not isnan(vv)) else None for vv in v]

    return data_payload


def is_nested_dataframe(df: pd.DataFrame, column: str) -> bool:
    """Checks if a dataframe contains cell entries which are series.

    Args:
        df (pd.DataFrame): Input dataframe.
        column (str): A column to check in the input dataframe.
    Returns:
        bool: True if the column contains pandas series, False otherwise.
    """
    return isinstance(df.iloc[0][column], pd.Series)


def check_nested_lengths(df: pd.DataFrame, columns: List[str]):
    """Check that each row contains series of the same length

    Args:
        df (pd.DataFrame): Input dataframe, assumed to contain nested series.
        columns (List[str]): The columns of the dataframe to consider.

    Raises:
        ValueError: Raised if the dataframe does not contain rows where each series is
        not equal length.
    """
    if len(columns) == 1:
        return

    l = df[columns].map(len).values
    if not np.all(np.isclose(l, np.tile(l[:, :1], (1, l.shape[1])))):
        raise ValueError("Input dataframe contains rows with series that are not equal length.")



import os
import tempfile
from pathlib import Path

from tsfminference.dirutil import resolve_model_path


def test_resolve_model_path():
    with tempfile.TemporaryDirectory() as dir1:
        with tempfile.TemporaryDirectory() as dir2:
            dirpath = f"{dir1}:{dir2}"
            os.mkdir(Path(dir1) / "amodel")
            assert resolve_model_path(dirpath, "amodel") == Path(dir1) / "amodel"
            assert resolve_model_path(dirpath, "foobar") is None
            assert resolve_model_path("fzbatt:zap", "amodel") is None
            os.mkdir(Path(dir2) / "anewmodel")
            assert resolve_model_path(dirpath, "anewmodel") == Path(dir2) / "anewmodel"



# Copyright contributors to the TSFM project
#
"""TSPulse model configuration"""

from typing import List, Optional, Union

from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging


logger = logging.get_logger(__name__)

TSPULSE_PRETRAINED_CONFIG_ARCHIVE_MAP = {}


class TSPulseConfig(PretrainedConfig):
    r"""
    This is the configuration class to store the configuration of a [`TSPulseModel`]. It is used to instantiate a
    TSPulse model according to the specified arguments, defining the model architecture. Instantiating a
    configuration with the defaults will yield a similar configuration to that of the TSPulse {} architecture.

    Configuration objects inherit from [`PretrainedConfig`] and can be used to control the model outputs. Read the
    documentation from [`PretrainedConfig`] for more detailed information.


    Args:
        context_length (`int`, *optional*, defaults to 64):
            The context/history length for the input sequence.
        patch_length (`int`, *optional*, defaults to 8):
            The patch length for the input sequence.
        num_input_channels (`int`, *optional*, defaults to 1):
            Number of input variates. For univariate series, set to 1.
        patch_stride (`int`, *optional*, defaults to 8):
            Stride value for patching. Equal to patch_length yields non-overlapping patches.
        d_model (`int`, *optional*, defaults to 16):
            Hidden feature dimension of the model.
        expansion_factor (`int`, *optional*, defaults to 2):
            Expansion ratio for MLP layers. Controls MLP capacity.
        num_layers (`int`, *optional*, defaults to 3):
            Number of backbone layers in the encoder.
        dropout (`float`, *optional*, defaults to 0.2):
            Dropout probability used across backbone layers.
        mode (`str`, *optional*, defaults to `"common_channel"`):
            Channel mixing mode. Choose from `"common_channel"` or `"mix_channel"`.
        gated_attn (`bool`, *optional*, defaults to `True`):
            Whether to use gated attention inside mixer blocks.
        norm_mlp (`str`, *optional*, defaults to `"LayerNorm"`):
            Normalization strategy inside MLP blocks. `"LayerNorm"` or `"BatchNorm"`.
        self_attn (`bool`, *optional*, defaults to `False`):
            Enable tiny self-attention over patches. Typically, not recommended.
        self_attn_heads (`int`, *optional*, defaults to 1):
            Number of heads for self-attention, used if `self_attn=True`.
        use_positional_encoding (`bool`, *optional*, defaults to `False`):
            Enable positional encodings for self-attention layers.
        positional_encoding_type (`str`, *optional*, defaults to `"sincos"`):
            Type of positional encoding: `"random"` or `"sincos"`.
        scaling (`str` or `bool`, *optional*, defaults to `"revin"`):
            Local scaling strategy. `"mean"`, `"std"`, `"revin"` or `None`. Recommended to always use `"revin"`
            for improved performance.
        loss (`str`, *optional*, defaults to `"mse"`):
            Loss function to apply, e.g., `"mse"`, `"mae", "cross_entropy". For classification task,
            set loss as `"cross_entropy"`.
        init_std (`float`, *optional*, defaults to 0.02):
            Std deviation for truncated normal weight initialization.
        post_init (`bool`, *optional*, defaults to `False`):
            Whether to use Huggingface-style custom initialization.
        norm_eps (`float`, *optional*, defaults to 1e-5):
            Epsilon for normalization layers for numerical stability.
        head_dropout (`float`, *optional*, defaults to 0.2):
            Dropout for head layers.
        reconstruction_type (`str`, *optional*, defaults to `"patchwise"`):
            Type of reconstruction - `"patchwise"` or `"full"`.
        decoder_num_layers (`int`, *optional*, defaults to 3):
            Number of decoder layers.
        decoder_d_model (`int`, *optional*, defaults to 16):
            Hidden size for decoder layers.
        decoder_mode (`str`, *optional*, defaults to `"common_channel"`):
            Channel mixing strategy for decoder.
        num_patches_layerwise_scale (`List[float]`, *optional*):
            List of patch scaling factors per layer in encoder. Recommended to set to None
        num_channels_layerwise_scale (`List[float]`, *optional*):
            List of channel scaling factors per layer in encoder. Recommended to set to None
        d_model_layerwise_scale (`List[float]`, *optional*):
            List of d_model scaling factors per layer in encoder.  Recommended to set to None
        decoder_num_patches_layerwise_scale (`List[float]`, *optional*):
            List of patch expansion per layer in decoder.  Recommended to set to None
        decoder_num_channels_layerwise_scale (`List[float]`, *optional*):
            List of channel expansion per layer in decoder.  Recommended to set to None
        decoder_d_model_layerwise_scale (`List[float]`, *optional*):
            List of d_model expansion per layer in decoder.  Recommended to set to None
        num_targets (`int`, *optional*, defaults to 3):
            Refers to the number of labels in a class for a classification task.
        output_range (`List[int]`, *optional*):
            Output range to set. Recommended to set to None
        head_aggregation (`str`, *optional*):
            Aggregation strategy for the classification head. Recommended to set to None.
        head_aggregation_dim (`str`, *optional*, defaults to `"patch"`):
            Aggregation dimension for the classification head. `"patch"` or `"channel"`.
        init_linear (`str`, *optional*, defaults to `"pytorch"`):
            Strategy for linear layer initialization. <todo>
        init_embed (`str`, *optional*, defaults to `"pytorch"`):
            Strategy for embedding layer initialization.
        mask_ratio (`float`, *optional*):
            Mask ratio during training.
        mask_type (`str`, *optional*, defaults to `"block"`):
            Type of masking strategy to apply. Supported options:

            - `"block"`: Applies patch-level masking (contiguous chunks).
            - `"user"`: Applies a user-defined mask passed via `past_observed_mask` in the forward call.
            - `"hybrid"`: Combines both point-level and patch-level masking.
            - `"var_hybrid"`: Similar to `"hybrid"`, but uses a variable mask ratio randomly sampled in the range `[0, mask_ratio]`.
            - `"random"`: Applies masking to randomly selected individual time-points (point masking).
        mask_block_length (`int`, *optional*, defaults to 8):
            Block length for masking if `mask_type` is `"block"`.
        loss_apply_mode (`str`, *optional*, defaults to `"mask"`):
            Specifies how the loss should be applied. Supported options:

            - `"mask"`: Apply loss only on the masked positions.
            - `"full"`: Apply loss across the entire output sequence.
            - `"mask_and_full"`: Apply both masked-point loss and full-sequence loss together.
        use_learnable_mask_token (`bool`, *optional*, defaults to `True`):
            Whether to use a learnable token to replace masked inputs.
        minimum_scale (`float`, *optional*, defaults to 1e-5):
            Minimum value used in scaling operations.
        categorical_vocab_size_list (`List[int]`, *optional*):
            Vocabulary sizes for categorical input features. Disabled currently.
        fuse_fft (`bool`, *optional*, defaults to `False`):
            Whether to use FFT-based feature fusion.
        fft_weight (`float`, *optional*, defaults to 1.0):
            Weight of FFT loss term.
        fft_prob_weight (`float`, *optional*, defaults to 1.0):
            Weight for probabilistic FFT loss term.
        fft_prob_length (`int`, *optional*):
            Length of signal to use in FFT probabilistic loss.
        fft_prob_mode (`str`, *optional*, defaults to `"plain"`):
            Mode of FFT probability loss. Allowed values `"plain"` and `"log"`
        enable_fft_prob_loss (`bool`, *optional*, defaults to `True`):
            Whether to enable FFT probability loss.
        fft_original_signal_loss_weight (`float`, *optional*, defaults to 1.0):
            Weight for original TS reconstruction signal loss from FFT branch.
        fft_time_add_forecasting_pt_loss (`bool`, *optional*, defaults to `False`):
            Whether to add forecasting point loss in time domain.
        fft_time_add_forecasting_pt_loss_weight (`float`, *optional*, defaults to 1.0):
            Weight for forecasting point loss.
        prediction_length (`int`, *optional*, defaults to 8):
            Number of future points to forecast, if forecast head is enabled.
        channel_consistent_masking (`bool`, *optional*, defaults to `True`):
            Whether masking is consistent across channels.
        hydra_class_head (`List[int]`, *optional*):
            Number of classes per task in Hydra head. Recommended to set to None
        hydra_class_attention (`bool`, *optional*, defaults to `True`):
            Whether to apply attention in classification hydra head.
        free_channel_flow (`bool`, *optional*, defaults to `True`):
            Enables free gradient flow in channel mixer by initializing based on `"channel_mix_init"` param
        patch_register_tokens (`int`, *optional*, defaults to 8):
            Number of patch register tokens to add.
        channel_register_tokens (`int`, *optional*):
            Number of channel register tokens. Recommended to set to None
        fft_time_consistent_masking (`bool`, *optional*, defaults to `True`):
            If True, masked time-series is used for FFT. If False, Masking is enabled seperately in FFT space.
            Recommended to set to True.
        fft_mask_ratio (`float`, *optional*):
            Masking ratio for FFT components.
        fft_mask_strategy (`str`, *optional*, defaults to `"magnitude"`):
            Strategy to mask FFT components - `"magnitude"` or `"random"`.
        fft_remove_component (`str`, *optional*, defaults to `"last"`):
            Which frequency component to remove in FFT to match to the shape requirements.
            Allowed values are `"last"` or `"dc"`. Recommended to set to `"last"`
        head_reduce_d_model (`int`, *optional*):
            Reduction dimension for classification head. if head_aggregation is None, set it to 1 or 2. Else, set it to None
        channel_virtual_expand_scale (`int`, *optional*, defaults to 1):
            Virtual expansion factor for channel mixing. Enable only in number of channels are very low (say 1 to 3)
        classification_mode (`str`, *optional*, defaults to `"full_embedding"`):
            Strategy to determine which part of the decoder embedding is used in the classification head. Supported options:

            - `"full_embedding"`: Use the entire embedding.
            - `"long_embedding"`: Use only the first `2N` patch embeddings (excluding register tokens), where `N` is the number of input patches.
            - `"short_embedding"`: Use only the register token embeddings.
            - `"fft_embedding"`: Use only the patch embeddings from positions `N+1` to `2N`.
            - `"time_embedding"`: Use only the patch embeddings from positions `0` to `N`.
        head_reduce_channels (`int`, *optional*):
            Channel reduction parameter in classification head. Recommended to set to None
        head_attention (`bool`, *optional*, defaults to `False`):
            Whether to use attention in final head layer. Recommended to set to False, as we enable it in decoder.
        gated_attention_activation (`str`, *optional*, defaults to `"softmax"`):
            Activation function for gated attention. `"softmax"` or `"sigmoid"`.
        head_gated_attention_activation (`str`, *optional*, defaults to `"softmax"`):
            Activation for gated attention in the head. `"softmax"` or `"sigmoid"`.
        register_mixer_layers (`int`, *optional*):
            Number of mixer layerss to use in the mini-decoder for fft-prob and forecast predictions. If None,
            no decoder used.
        reconstruction_loss_weight (`int`, *optional*, defaults to 1):
            Weight for the reconstruction loss.
        masked_reconstruction_loss_weight (`int`, *optional*, defaults to 1):
            Weight for the masked reconstruction loss.
        channel_mix_init (`str`, *optional*, defaults to `"identity"`):
            Initialization strategy for channel mixing - `"identity"` or `"zero"`. Recommended to set to identity
        batch_aware_masking (`bool`, *optional*, defaults to `False`):
            If enabled, some samples in batches are time masked while others are fft masked. Recommended to set to False
        disable_mask_in_classification_eval (`bool`, *optional*, defaults to `True`):
            Whether to disable masking during evaluation in classification.
        num_full_patches_for_hybrid_mask (`int`, *optional*, defaults to 2):
            Number of full patches to mask for `"hybrid"` mask type.
        full_patch_mask_percentage (`float`, *optional*, defaults to 0.2):
            Percentage of full patches to mask in `"var_hybrid"` mask type.
        revin_affine (`bool`, *optional*, defaults to `True`):
            Whether to use affine RevIN transformation.
        data_actual_context_length (`int`, *optional*):
            Original context length before padding or sampling. Recommended to set to None. Not allowed.
        fft_applied_on (`str`, *optional*, defaults to `"scaled_ts"`):
            FFT applied on scaled or raw time-series: `"raw_ts"` or `"scaled_ts"`. Recommended to set to scaled_ts

    """

    model_type = "tspulse"
    attribute_map = {
        "hidden_size": "d_model",
        "num_hidden_layers": "num_layers",
    }

    def __init__(
        self,
        # Time series specific configuration
        context_length: int = 64,
        patch_length: int = 8,
        num_input_channels: int = 1,
        patch_stride: int = 8,
        # General model configuration
        d_model: int = 16,
        expansion_factor: int = 2,
        num_layers: int = 3,
        dropout: float = 0.2,
        mode: str = "common_channel",
        gated_attn: bool = True,
        norm_mlp: str = "LayerNorm",
        self_attn: bool = False,
        self_attn_heads: int = 1,
        use_positional_encoding: bool = False,
        positional_encoding_type: str = "sincos",
        scaling: Optional[Union[str, bool]] = "revin",
        loss: str = "mse",
        init_std: float = 0.02,
        post_init: bool = False,
        norm_eps: float = 1e-5,
        head_dropout: float = 0.2,
        reconstruction_type: str = "patchwise",
        decoder_num_layers: int = 3,
        decoder_d_model: int = 16,
        decoder_mode: str = "common_channel",
        num_patches_layerwise_scale: Optional[List[float]] = None,
        num_channels_layerwise_scale: Optional[List[float]] = None,
        d_model_layerwise_scale: Optional[List[float]] = None,
        decoder_num_patches_layerwise_scale: Optional[List[float]] = None,
        decoder_num_channels_layerwise_scale: Optional[List[float]] = None,
        decoder_d_model_layerwise_scale: Optional[List[float]] = None,
        num_targets: int = 3,
        output_range: Optional[list] = None,
        head_aggregation: Optional[str] = None,
        head_aggregation_dim: str = "patch",
        init_linear: str = "pytorch",
        init_embed: str = "pytorch",
        mask_ratio: Optional[float] = None,
        mask_type: str = "block",  # block, point
        mask_block_length: int = 8,
        loss_apply_mode: str = "mask",
        use_learnable_mask_token: bool = True,
        minimum_scale: float = 1e-5,
        categorical_vocab_size_list: Optional[list] = None,
        fuse_fft: Optional[bool] = False,
        fft_weight: float = 1,
        fft_prob_weight: float = 1,
        fft_prob_length: Optional[int] = None,
        fft_prob_mode: str = "plain",
        enable_fft_prob_loss: bool = True,
        fft_original_signal_loss_weight: float = 1,
        fft_time_add_forecasting_pt_loss: bool = False,
        fft_time_add_forecasting_pt_loss_weight: float = 1,
        prediction_length: Optional[int] = 96,
        channel_consistent_masking: Optional[bool] = True,
        hydra_class_head: Optional[list] = None,
        hydra_class_attention: Optional[bool] = True,
        free_channel_flow: bool = True,
        patch_register_tokens: Optional[int] = 8,
        channel_register_tokens: Optional[int] = None,
        fft_time_consistent_masking: Optional[bool] = True,
        fft_mask_ratio: Optional[float] = None,
        fft_mask_strategy: Optional[str] = "magnitude",
        fft_remove_component: Optional[str] = "last",
        head_reduce_d_model: Optional[int] = None,
        channel_virtual_expand_scale: Optional[int] = 1,
        classification_mode: Optional[str] = "full_embedding",
        head_reduce_channels: Optional[int] = None,
        head_attention: Optional[bool] = False,
        gated_attention_activation: Optional[str] = "softmax",
        head_gated_attention_activation: str = "softmax",
        register_mixer_layers: Optional[int] = None,
        reconstruction_loss_weight: Optional[int] = 1,
        masked_reconstruction_loss_weight: Optional[int] = 1,
        channel_mix_init: Optional[str] = "identity",
        batch_aware_masking: Optional[bool] = False,
        disable_mask_in_classification_eval: Optional[bool] = True,
        num_full_patches_for_hybrid_mask: Optional[int] = 2,
        full_patch_mask_percentage: Optional[float] = 0.2,
        revin_affine: Optional[bool] = True,
        data_actual_context_length: Optional[int] = None,
        fft_applied_on: Optional[str] = "scaled_ts",  # "raw_ts",  # "scaled_ts",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.d_model = d_model
        self.decoder_d_model = decoder_d_model
        self.patch_stride = patch_stride
        self.num_input_channels = num_input_channels
        self.context_length = context_length
        self.patch_length = patch_length
        self.expansion_factor = expansion_factor

        self.num_full_patches_for_hybrid_mask = num_full_patches_for_hybrid_mask
        self.reconstruction_loss_weight = reconstruction_loss_weight
        self.masked_reconstruction_loss_weight = masked_reconstruction_loss_weight
        self.num_layers = num_layers
        self.dropout = dropout
        self.mode = mode
        self.gated_attn = gated_attn
        self.norm_mlp = norm_mlp
        self.scaling = scaling
        self.head_dropout = head_dropout
        self.patch_last = True
        self.use_positional_encoding = use_positional_encoding
        self.positional_encoding_type = positional_encoding_type
        self.self_attn = self_attn
        self.self_attn_heads = self_attn_heads
        self.init_std = init_std
        self.post_init = post_init
        self.enable_fft_prob_loss = enable_fft_prob_loss
        self.loss = loss
        self.norm_eps = norm_eps
        self.min_allowed_range = 4
        self.decoder_num_layers = decoder_num_layers
        self.decoder_mode = decoder_mode
        self.reconstruction_type = reconstruction_type
        self.fft_prob_weight = fft_prob_weight
        self.gated_attention_activation = gated_attention_activation
        self.register_mixer_layers = register_mixer_layers
        self.fft_prob_mode = fft_prob_mode
        self.classification_mode = classification_mode
        self.head_gated_attention_activation = head_gated_attention_activation
        self.fft_remove_component = fft_remove_component
        self.use_learnable_mask_token = use_learnable_mask_token
        self.channel_consistent_masking = channel_consistent_masking
        self.fft_prob_length = fft_prob_length
        self.num_patches_layerwise_scale = num_patches_layerwise_scale
        self.num_channels_layerwise_scale = num_channels_layerwise_scale
        self.d_model_layerwise_scale = d_model_layerwise_scale
        self.channel_mix_init = channel_mix_init
        self.decoder_d_model_layerwise_scale = decoder_d_model_layerwise_scale
        self.decoder_num_channels_layerwise_scale = decoder_num_channels_layerwise_scale
        self.decoder_num_patches_layerwise_scale = decoder_num_patches_layerwise_scale
        self.num_targets = num_targets
        self.output_range = output_range
        self.head_aggregation = head_aggregation
        self.head_aggregation_dim = head_aggregation_dim

        self.data_actual_context_length = data_actual_context_length

        self.revin_affine = revin_affine

        self.hydra_class_attention = hydra_class_attention
        self.free_channel_flow = free_channel_flow
        self.head_attention = head_attention
        self.fft_original_signal_loss_weight = fft_original_signal_loss_weight

        # init
        self.init_linear = init_linear
        self.init_embed = init_embed
        self.head_reduce_d_model = head_reduce_d_model
        self.channel_virtual_expand_scale = channel_virtual_expand_scale

        self.init_processing = False

        self.disable_mask_in_classification_eval = disable_mask_in_classification_eval
        self.fuse_fft = fuse_fft
        self.fft_weight = fft_weight
        self.mask_ratio = mask_ratio
        self.mask_block_length = mask_block_length
        self.mask_type = mask_type
        self.loss_apply_mode = loss_apply_mode

        self.head_reduce_channels = head_reduce_channels

        self.minimum_scale = minimum_scale
        self.categorical_vocab_size_list = categorical_vocab_size_list

        self.patch_register_tokens = patch_register_tokens

        self.channel_register_tokens = channel_register_tokens

        self.full_patch_mask_percentage = full_patch_mask_percentage

        self.fft_time_add_forecasting_pt_loss = fft_time_add_forecasting_pt_loss
        self.fft_time_add_forecasting_pt_loss_weight = fft_time_add_forecasting_pt_loss_weight

        self.batch_aware_masking = batch_aware_masking

        self.fft_applied_on = fft_applied_on

        self.fft_time_consistent_masking = fft_time_consistent_masking
        self.hydra_class_head = hydra_class_head

        self.fft_mask_ratio = fft_mask_ratio
        self.fft_mask_strategy = fft_mask_strategy

        self.prediction_length = prediction_length

    def set_scale(self, scale_param, actual_param, base_param, no_levels):
        if self._check_one_or_none(getattr(self, scale_param)):
            setattr(self, scale_param, [1] * no_levels)
            setattr(self, actual_param, [getattr(self, base_param)] * no_levels)
        else:
            setattr(
                self,
                actual_param,
                [int(getattr(self, base_param) * float(i)) for i in getattr(self, scale_param)],
            )

        for i in getattr(self, actual_param):
            if i < self.min_allowed_range and i != getattr(self, base_param):
                raise ValueError(
                    "Too much compression beyond level %s for param: %s, %s"
                    % (self.min_allowed_range, scale_param, getattr(self, actual_param))
                )

    def _check_one_or_none(self, param):
        if param is None:
            return True
        elif all(element == 1 for element in param):
            return True
        return False

    def check_and_init_preprocessing(self, task="reconstruction"):
        self.init_processing = True
        # print("Number of channels:", self.num_input_channels)

        if self.data_actual_context_length is not None:
            raise Exception("Set data_actual_context_length to None. Variable length input is not allowed currently.")

        if self.channel_register_tokens is not None:
            raise Exception("channel_register_tokens is not allowed. Set channel_register_tokens to None")
        if self.fuse_fft and self.context_length % 2 != 0:
            raise Exception("context length needs to be divisible by 2 when fuse_fft is True")

        if self.patch_length != self.patch_stride:
            raise Exception("Overlapping patches are not allowed currently..")

        if self.mask_ratio and self.mask_ratio > 0:
            if self.mask_block_length != self.patch_length:
                logger.warning("Please be aware that Mask block length is set different from the patch length")

        # if self.total_embedding_size is None:
        #     raise Exception("total_embedding_size cannot be None")

        if self.patch_register_tokens is None:
            raise Exception("Patch register tokens cannot be None..")

        if not hasattr(self, "num_patches"):
            self.num_patches = (
                max(self.context_length, self.patch_length) - self.patch_length
            ) // self.patch_stride + 1

            if self.fuse_fft:
                self.num_patches = self.num_patches * 2

        if self.mode == "common_channel":
            if not self._check_one_or_none(self.num_channels_layerwise_scale) or not self._check_one_or_none(
                self.decoder_num_channels_layerwise_scale
            ):
                logger.warning("Channel Compression not allowed when mode is common_channel. Setting it not None")
                self.num_channels_layerwise_scale = None
                self.decoder_num_channels_layerwise_scale = None

        if task in ["classification"]:
            if not self._check_one_or_none(self.num_channels_layerwise_scale):
                logger.warning("Channel Compression is not allowed in encoder for classification. Setting it to None")
                self.num_channels_layerwise_scale = None

        self.set_scale(
            scale_param="num_patches_layerwise_scale",
            actual_param="num_patches_layerwise",
            base_param="num_patches",
            no_levels=self.num_layers,
        )

        self.set_scale(
            scale_param="num_channels_layerwise_scale",
            actual_param="num_channels_layerwise",
            base_param="num_input_channels",
            no_levels=self.num_layers,
        )

        self.set_scale(
            scale_param="d_model_layerwise_scale",
            actual_param="d_model_layerwise",
            base_param="d_model",
            no_levels=self.num_layers,
        )

        self.set_scale(
            scale_param="decoder_num_patches_layerwise_scale",
            actual_param="decoder_num_patches_layerwise",
            base_param="num_patches",
            no_levels=self.decoder_num_layers,
        )

        self.set_scale(
            scale_param="decoder_num_channels_layerwise_scale",
            actual_param="decoder_num_channels_layerwise",
            base_param="num_input_channels",
            no_levels=self.decoder_num_layers,
        )

        self.set_scale(
            scale_param="decoder_d_model_layerwise_scale",
            actual_param="decoder_d_model_layerwise",
            base_param="decoder_d_model",
            no_levels=self.decoder_num_layers,
        )



# Copyright contributors to the TSFM project
#

"""Tests learning rate finder utility"""

import numpy as np
from transformers import set_seed

from tsfm_public.toolkit.dataset import ForecastDFDataset
from tsfm_public.toolkit.lr_finder import optimal_lr_finder
from tsfm_public.toolkit.util import select_by_index


def test_lr_finder(ttm_base_model, etth_data_base):
    set_seed(42)
    model = ttm_base_model

    context_length = 512
    prediction_length = 96
    timestamp_column = "date"
    id_columns = []
    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]

    data = etth_data_base.copy()

    train_end_index = 12 * 30 * 24 + 8 * 30 * 24
    train_start_index = train_end_index - context_length - prediction_length

    train_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=train_start_index,
        end_index=train_end_index,
    )

    train_dataset = ForecastDFDataset(
        train_data,
        timestamp_column=timestamp_column,
        target_columns=target_columns,
        prediction_length=prediction_length,
        context_length=context_length,
    )

    learning_rate, finetune_forecast_model = optimal_lr_finder(
        model, train_dataset, device="cpu", batch_size=32, num_iter=10
    )

    np.testing.assert_allclose(learning_rate, 0.00035938136638046257)

    learning_rate, finetune_forecast_model = optimal_lr_finder(
        model, train_dataset, device="mps", batch_size=32, num_iter=10
    )



# Copyright contributors to the TSFM project
#

"""Tests data handling functions"""

import numpy as np

from tsfm_public import load_dataset


def test_load_dataset():
    dset_train, dset_valid, dset_test = load_dataset(
        dataset_name="etth1",
        context_length=512,
        forecast_length=96,
        fewshot_fraction=1.0,
        dataset_path="https://raw.githubusercontent.com/zhouhaoyi/ETDataset/main/ETT-small/ETTh1.csv",
    )

    np.testing.assert_allclose([len(x) for x in [dset_train, dset_valid, dset_test]], [8033, 2785, 2785])



# Copyright contributors to the TSFM project
#

"""Utilities for testing"""

from itertools import chain, repeat


def nreps(iterable, n):
    "Returns each element in the sequence repeated n times."
    return chain.from_iterable((repeat(i, n) for i in iterable))



# Copyright contributors to the TSFM project
#

try:
    # Local
    from ._version import __version__, __version_tuple__  # noqa: F401 # unused import
except ImportError:
    __version__ = "unknown"
    __version_tuple__ = (0, 0, __version__)



# Copyright contributors to the TSFM project
#
"""Core code for the demo app"""

import logging
import re

import numpy as np
import pandas as pd
import streamlit as st

from tsfmhfdemos.neurips.backends.v1 import GLOBAL_CONFIG, model_util


logging.basicConfig(level=logging.INFO)

# **********************************************************
# There should be no need to edit anything below
# unless you are altering the content of the
# UI itself
# **********************************************************

logging.info(GLOBAL_CONFIG)

MODEL_DOCS = GLOBAL_CONFIG["MODEL_DOCS"]
create_figure = model_util.create_figure
# provides the forecast for a pretrained model
forecast = model_util.forecast

DATASETS = GLOBAL_CONFIG["DATASETS"]
MODELS = GLOBAL_CONFIG["MODELS"]
INFERENCE_APPROACHES = GLOBAL_CONFIG["INFERENCE_APPROACHES"]
METRICS = GLOBAL_CONFIG["INFERENCE_METRICS"]


def tsforecasting_with_fmdls():
    st.set_page_config(
        page_title=GLOBAL_CONFIG["title"],
        page_icon="🔮",
        layout="wide",
        initial_sidebar_state="expanded",
    )

    st.title(GLOBAL_CONFIG["title"])
    st.write("<style>div.block-container{padding-top:2rem;}</style>", unsafe_allow_html=True)

    st.write(GLOBAL_CONFIG["intro"])

    # required_cols = ["ds", "y"]

    with st.sidebar:
        st.subheader("Select the pre-trained model to use.")
        model_name = st.radio(
            "Model:",
            MODELS.keys(),
        )

    with st.sidebar:
        st.subheader("Select which dataset to use for tuning and testing.")
        dataset_name = st.radio(
            "Dataset:",
            DATASETS.keys(),
        )

    with st.sidebar:
        st.subheader("Select the tuning approach.")
        approach_name = st.radio(
            "Tuning Approach:",
            INFERENCE_APPROACHES.keys(),
        )

        dataset_meta = DATASETS[dataset_name]
        model_meta = MODELS[model_name]
        approach_meta = INFERENCE_APPROACHES[approach_name]

    # tabs
    tab_forecast, tab_leaderboard, tab_docs = st.tabs(
        [
            ":chart: Forecast",
            ":runner: Leader Board",
            ":information_source: Documentation",
        ]
    )

    with tab_forecast:
        # ### Dataframe of selected data
        # st.json(GLOBAL_CONFIG)

        st.header(dataset_name)

        # obtain forecast
        # forecasts = model_util.forecast(**dataset_meta, **model_meta)
        # plot forecast and ground true
        # num_cols = 1
        # columns = st.columns(num_cols)

        col1, col2 = st.columns([0.7, 0.3])

        with col1:
            st.subheader("Forecast results")
            for idx, channel in enumerate(dataset_meta["channel_plots"]):
                # col = columns[idx % num_cols]
                st.plotly_chart(
                    model_util.create_figure(**dataset_meta, **model_meta, **approach_meta, channel=channel),
                    use_container_width=True,
                    fig_size=(1600, 200),
                )

        with col2:
            st.subheader("Performance")
            df_perf = model_util.get_performance(metrics=METRICS, **dataset_meta, **model_meta, **approach_meta)

            df_perf_styled = df_perf.style.set_table_styles(
                [
                    {"selector": "th", "props": "background-color: whitesmoke;"},
                ]
            ).format(precision=3)  # .style.hide(axis="index")
            st.write(df_perf_styled.to_html(), unsafe_allow_html=True)
            st.write("")

        # st.dataframe(df_perf)
        # add output of results

        st.subheader("Dataset")
        df = model_util.csv_to_df(dataset_meta)
        st.dataframe(df)

    with tab_leaderboard:
        st.subheader("Current Leader Board")
        # leaderboard = pd.DataFrame(
        #     np.random.randn(len(INFERENCE_APPROACHES), len(METRICS)),
        #     columns=METRICS,
        #     index=INFERENCE_APPROACHES,
        # )

        table_source = r"""
\begin{tabular}{cc|c|cc|cc|cc|cc|cc|ccc}
\cline{2-15}
&\multicolumn{2}{c|}{Models} & \multicolumn{2}{c}{\textbf{\citsm-Best}} & \multicolumn{2}{c|}{DLinear} & \multicolumn{2}{c|}{PatchTST}& \multicolumn{2}{c|}{FEDformer}& \multicolumn{2}{c|}{Autoformer}& \multicolumn{2}{c}{Informer} \\
\cline{2-15}
&\multicolumn{2}{c|}{Metric}&MSE&MAE&MSE&MAE&MSE&MAE&MSE&MAE&MSE&MAE&MSE&MAE\\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{ETTH1}} & 96 & \textbf{0.368$\pm$0.001} & \textbf{0.398$\pm$0.001} & 0.375 & \uline{0.399} & \uline{0.370} & 0.400 & 0.376 & 0.419 & 0.449 & 0.459 & 0.865 & 0.713 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.399$\pm$0.002} & \uline{0.418$\pm$0.001} & \uline{0.405} & \textbf{0.416} & 0.413 & 0.429 & 0.420 & 0.448 & 0.500 & 0.482 & 1.008 & 0.792 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.421$\pm$0.004} & \textbf{0.436$\pm$0.003} & 0.439 & 0.443 & \uline{0.422} & \uline{0.440} & 0.459 & 0.465 & 0.521 & 0.496 & 1.107 & 0.809 \\
&\multicolumn{1}{c|}{} & 720 & \textbf{0.444$\pm$0.003} & \textbf{0.467$\pm$0.002} & 0.472 & 0.490 & \uline{0.447} & \uline{0.468} & 0.506 & 0.507 & 0.514 & 0.512 & 1.181 & 0.865 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{ETTH2}} & 96 & \uline{0.276$\pm$0.006} & \textbf{0.337$\pm$0.003} & 0.289 & \uline{0.353} & \textbf{0.274} & \textbf{0.337} & 0.346 & 0.388 & 0.358 & 0.397 & 3.755 & 1.525 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.330$\pm$0.003} & \textbf{0.374$\pm$0.001} & 0.383 & 0.418 & \uline{0.341} & \uline{0.382} & 0.429 & 0.439 & 0.456 & 0.452 & 5.602 & 1.931 \\
&\multicolumn{1}{c|}{} & 336 & \uline{0.357$\pm$0.001} & \uline{0.401$\pm$0.002} & 0.448 & 0.465 & \textbf{0.329} & \textbf{0.384} & 0.496 & 0.487 & 0.482 & 0.486 & 4.721 & 1.835 \\
&\multicolumn{1}{c|}{} & 720 & \uline{0.395$\pm$0.003} & \uline{0.436$\pm$0.003} & 0.605 & 0.551 & \textbf{0.379} & \textbf{0.422} & 0.463 & 0.474 & 0.515 & 0.511 & 3.647 & 1.625 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{ETTM1}} & 96 & \textbf{0.291$\pm$0.002} & \uline{0.346$\pm$0.002} & 0.299 & \textbf{0.343} & \uline{0.293} & \uline{0.346} & 0.379 & 0.419 & 0.505 & 0.475 & 0.672 & 0.571 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.333$\pm$0.002} & \uline{0.369$\pm$0.002} & \uline{0.335} & \textbf{0.365} & \textbf{0.333} & 0.370 & 0.426 & 0.441 & 0.553 & 0.496 & 0.795 & 0.669 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.365$\pm$0.005} & \textbf{0.385$\pm$0.004} & \uline{0.369} & \uline{0.386} & \uline{0.369} & 0.392 & 0.445 & 0.459 & 0.621 & 0.537 & 1.212 & 0.871 \\
&\multicolumn{1}{c|}{} & 720 & \textbf{0.416$\pm$0.002} & \textbf{0.413$\pm$0.001} & \uline{0.425} & 0.421 & \textbf{0.416} & \uline{0.420} & 0.543 & 0.490 & 0.671 & 0.561 & 1.166 & 0.823 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{ETTM2}} & 96 & \textbf{0.164$\pm$0.002} & \textbf{0.255$\pm$0.002} & 0.167 & 0.260 & \uline{0.166} & \uline{0.256} & 0.203 & 0.287 & 0.255 & 0.339 & 0.365 & 0.453 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.219$\pm$0.002} & \textbf{0.293$\pm$0.002} & 0.224 & 0.303 & \uline{0.223} & \uline{0.296} & 0.269 & 0.328 & 0.281 & 0.340 & 0.533 & 0.563 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.273$\pm$0.003} & \textbf{0.329$\pm$0.003} & 0.281 & \uline{0.342} & \uline{0.274} & \textbf{0.329} & 0.325 & 0.366 & 0.339 & 0.372 & 1.363 & 0.887 \\
&\multicolumn{1}{c|}{} & 720 & \textbf{0.358$\pm$0.002} & \textbf{0.380$\pm$0.001} & 0.397 & 0.421 & \uline{0.362} & \uline{0.385} & 0.421 & 0.415 & 0.433 & 0.432 & 3.379 & 1.338 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{Electricity}} & 96 & \textbf{0.129$\pm$1e-4} & \uline{0.224$\pm$0.001} & \uline{0.140} & 0.237 & \textbf{0.129} & \textbf{0.222} & 0.193 & 0.308 & 0.201 & 0.317 & 0.274 & 0.368 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.146$\pm$0.001} & \uline{0.242$\pm$1e-4} & 0.153 & 0.249 & \uline{0.147} & \textbf{0.240} & 0.201 & 0.315 & 0.222 & 0.334 & 0.296 & 0.386 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.158$\pm$0.001} & \textbf{0.256$\pm$0.001} & 0.169 & 0.267 & \uline{0.163} & \uline{0.259} & 0.214 & 0.329 & 0.231 & 0.338 & 0.300 & 0.394 \\
&\multicolumn{1}{c|}{} & 720 & \textbf{0.186$\pm$0.001} & \textbf{0.282$\pm$0.001} & 0.203 & 0.301 & \uline{0.197} & \uline{0.290} & 0.246 & 0.355 & 0.254 & 0.361 & 0.373 & 0.439 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{Traffic}} & 96 & \textbf{0.356$\pm$0.002} & \textbf{0.248$\pm$0.002} & 0.410 & 0.282 & \uline{0.360} & \uline{0.249} & 0.587 & 0.366 & 0.613 & 0.388 & 0.719 & 0.391 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.377$\pm$0.003} & \uline{0.257$\pm$0.002} & 0.423 & 0.287 & \uline{0.379} & \textbf{0.256} & 0.604 & 0.373 & 0.616 & 0.382 & 0.696 & 0.379 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.385$\pm$0.002} & \textbf{0.262$\pm$0.001} & 0.436 & 0.296 & \uline{0.392} & \uline{0.264} & 0.621 & 0.383 & 0.622 & 0.337 & 0.777 & 0.420 \\
&\multicolumn{1}{c|}{} & 720 & \textbf{0.424$\pm$0.001} & \textbf{0.283$\pm$0.001} & 0.466 & 0.315 & \uline{0.432} & \uline{0.286} & 0.626 & 0.382 & 0.660 & 0.408 & 0.864 & 0.472 \\
\cline{2-15}
&\multirow{4}*{\rotatebox{90}{Weather}} & 96 & \textbf{0.146$\pm$0.001} & \textbf{0.197$\pm$0.002} & 0.176 & 0.237 & \uline{0.149} & \uline{0.198} & 0.217 & 0.296 & 0.266 & 0.336 & 0.300 & 0.384 \\
&\multicolumn{1}{c|}{} & 192 & \textbf{0.191$\pm$0.001} & \textbf{0.240$\pm$0.001} & 0.220 & 0.282 & \uline{0.194} & \uline{0.241} & 0.276 & 0.336 & 0.307 & 0.367 & 0.598 & 0.544 \\
&\multicolumn{1}{c|}{} & 336 & \textbf{0.243$\pm$0.001} & \textbf{0.279$\pm$0.002} & 0.265 & 0.319 & \uline{0.245} & \uline{0.282} & 0.339 & 0.380 & 0.359 & 0.395 & 0.578 & 0.523 \\
&\multicolumn{1}{c|}{} & 720 & \uline{0.316$\pm$0.001} & \textbf{0.333$\pm$0.002} & 0.323 & 0.362 & \textbf{0.314} & \uline{0.334} & 0.403 & 0.428 & 0.419 & 0.428 & 1.059 & 0.741 \\
\cline{2-15}
% &\multicolumn{4}{c|}{\makecell{\textbf{\citsm-Best} \textbf{\% improvement}}}& \textbf{8\%} & \textbf{6.8\%}& \textbf{0.7\%} & \textbf{0.4\%} & \textbf{22.9\%} & \textbf{18.2\%} & \textbf{30.1\%} & \textbf{22.7\%} & \textbf{64\%} & \textbf{50.3\%} \\
% &\multicolumn{4}{c|}{\makecell{\textbf{\citsm-Best} \textbf{\% improvement (MSE)}}}& \multicolumn{2}{c}{\textbf{8\%}} & \multicolumn{2}{c}{ \textbf{0.7\%}}  & \multicolumn{2}{c}{\textbf{22.9\%}}  & \multicolumn{2}{c}{\textbf{30.1\%}}  & \multicolumn{2}{c}{\textbf{64\%}}  \\
&\multicolumn{4}{c|}{\makecell{\textbf{\citsm-Best} \textbf{\% improvement (MSE)}}}& \multicolumn{2}{c}{\textbf{8\%}} & \multicolumn{2}{c}{ \textbf{1\%}}  & \multicolumn{2}{c}{\textbf{23\%}}  & \multicolumn{2}{c}{\textbf{30\%}}  & \multicolumn{2}{c}{\textbf{64\%}}  \\
\cline{2-15}
\end{tabular}
"""

        out = re.sub(r"\\textbf{([^&]*)}", r"\1", table_source)
        out = re.sub(r"\\uline{([^&]*)}", r"\1", out)
        out = re.sub(r"\s*|\$\\pm\$[^&]*|\\cline{.*}", "", out)
        vals = np.array([r.split("&")[3:] for r in out.split(r"\\")[2:30]]).astype(float)

        leaderboard = pd.DataFrame(
            index=pd.MultiIndex.from_product(
                [
                    [
                        "ETTh1",
                        "ETTh2",
                        "ETTm1",
                        "ETTm2",
                        "Electricity",
                        "Traffic",
                        "Weather",
                    ],
                    [96, 192, 336, 720],
                ],
                names=["Dataset", "Prediction length"],
            ),
            columns=pd.MultiIndex.from_product(
                [
                    [
                        "PatchTSMixer",
                        "DLinear",
                        "PatchTST",
                        "FEDformer",
                        "Autoformer",
                        "Informer",
                    ],
                    ["MSE", "MAE"],
                ],
                names=["Model", "Metric"],
            ),
            data=vals,
        )

        def highlight_best(s, props=""):
            for metric in ["MSE", "MAE"]:
                mask = s.index.droplevel(0) == metric
                s[mask] = s[mask] == np.nanmin(s.values[mask])

            ret = np.where(s, props, "")
            return ret

        def highlight_second_best(s, props=""):
            for metric in ["MSE", "MAE"]:
                mask = s.index.droplevel(0) == metric
                v_ord = np.sort(s.values[mask])
                v_min = v_ord[0]
                v_second = v_ord[1]
                s[mask] = (s[mask] == v_second) & (v_min != v_second)

            ret = np.where(s, props, "")
            return ret

        # (df.style.apply(highlight_max, axis=0, props='background-color:green;', subset=['A','B'])
        #  .apply(highlight_min, axis=0, props='background-color:red;', subset=['A','B'])

        # st.dataframe(leaderboard, hide_index=False)

        leaderboard = leaderboard[
            [
                "PatchTSMixer",
                "PatchTST",
                "DLinear",
                "FEDformer",
                "Autoformer",
                "Informer",
            ]
        ]

        leaderboard_styled = (
            leaderboard.style.set_table_styles(
                [
                    {"selector": "th", "props": "background-color: whitesmoke;"},
                ]
            )
            .format(precision=3)
            .apply(highlight_best, axis=1, props="font-weight: bold;")
            .apply(highlight_second_best, axis=1, props="text-decoration: underline;")
        )

        st.write(leaderboard_styled.to_html(), unsafe_allow_html=True)
        # st.dataframe(leaderboard_styled)

        st.write("Source: https://arxiv.org/abs/2306.09364")

    with tab_docs:
        # drop [0] below if more than one tab
        tab_desc = st.tabs(["Description of the models"])[0]

        with tab_desc:
            for model_name in MODELS.keys():
                st.header(model_name)
                model_card_name = MODELS[model_name]["card"]
                st.subheader("Abstract")
                st.write(f"""{MODEL_DOCS[model_card_name]["Abstract"]}""")
                st.subheader("Model Architecture")
                st.image(**MODEL_DOCS[model_card_name]["figure"])
                st.write(f"""{MODEL_DOCS[model_card_name]["Model Architecture"]}""")
                # st.subheader("Secondary use")
                # st.write(f"""{MODEL_DOCS[model_card_name]['Secondary use']}""")
                # st.subheader("Limitations")
                # st.write(f"""{MODEL_DOCS[model_card_name]['Limitations']}""")
                # st.subheader("Training data")
                # st.write(f"""{MODEL_DOCS[model_card_name]['Training data']}""")
                st.subheader("BibTex/Citation Info")
                st.code(f"""{MODEL_DOCS[model_card_name]["Citation Info"]}""")


if __name__ == "__main__":
    tsforecasting_with_fmdls()



# Copyright contributors to the TSFM project
#
"""Utilities for plotting time series data"""

import inspect
import logging
import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import plotly.graph_objs as go
import torch
from IPython.display import Image
from plotly.subplots import make_subplots
from torch.utils.data.dataset import Dataset
from transformers import PreTrainedModel

from .time_series_preprocessor import create_timestamps


try:
    import matplotlib.pyplot as plt
    import seaborn as sns

    HAVE_SEABORN = True
except ImportError:
    HAVE_SEABORN = False


def explode_and_add_timestamps(df, value_column, timestamp_column, time_unit="h"):
    # create df with all time predictions
    df_exploded = df.explode(value_column, ignore_index=True)
    df_exploded["periods"] = df_exploded.groupby(timestamp_column).cumcount()
    df_exploded["offset"] = pd.to_timedelta(df_exploded["periods"], unit=time_unit)
    df_exploded["new_timestamp"] = df_exploded[timestamp_column] + df_exploded["offset"]
    df_exploded = df_exploded.drop(columns=["periods", "offset"])
    return df_exploded


def plot_line(x, y, plot_type: str, plot_id: str, data=None, fig=None, **kwargs):
    # plot data lines based on plot type
    if plot_type == "plotly":
        fig.add_trace(go.Scatter(x=x, y=y, name=plot_id, **kwargs))
    else:
        sns.lineplot(data=data, x=x, y=y, label=plot_id, **kwargs)


def plot_ts_forecasting(
    test_data: pd.DataFrame,
    forecast_data: pd.DataFrame,
    forecast_columns: list,
    timestamp_column: str,
    periodicity: str,
    prediction_length: int,
    context_length: int,
    plot_stride: int = None,
    num_predictions: int = None,
    plot_start: int = 0,
    plot_end: int = -1,
    title: str = "Forecasting plot",
    fig_size: tuple = (15, 8),
    plot_type: str = "seaborn",
    return_image: bool = True,
):
    """_summary_

    Args:
        test_data (pd.DataFrame): _description_
        forecast_data (pd.DataFrame): _description_
        forecast_columns (list): _description_
        timestamp_column (str): _description_
        periodicity (str): _description_
        prediction_length (int): _description_
        context_length (int): _description_
        plot_stride (int, optional): _description_. Defaults to None.
        plot_start (int, optional): _description_. Defaults to 0.
        plot_end (int, optional): _description_. Defaults to -1.
        title (str, optional): _description_. Defaults to "Forecasting plot".
        fig_size (tuple, optional): _description_. Defaults to (15, 8).
        plot_type (str, optional): _description_. Defaults to "seaborn".
        return_image (bool, optional): _description_. Defaults to True.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    test_data_updated = test_data  # .reset_index()
    # plot true data

    if not HAVE_SEABORN and plot_type == "seaborn":
        raise ValueError("Please install the seaborn package if seaborn plots are needed.")

    # if plot_start > len(test_data_updated):
    #     logging.warning(
    #         f"The start index is out of range. "
    #         f"Please choose the start index within the range of the test data to see the true values."
    #     )

    if not periodicity:
        logging.warning("Please specify periodicity.")

    if not context_length:
        logging.warning("Please specify context_length.")

    if not prediction_length:
        logging.warning("Please specify prediction_length.")

    plot_range = range(test_data_updated.shape[0])[plot_start:plot_end]

    # if plot_range[0] < context_length:
    #     logging.warning(
    #         f"To see the plot lines of the prediction and ground truth "
    #         f"choose the plot_start larger than {context_length}."
    #     )

    # plot_end = min(plot_end, len(test_data_updated))
    # plot_start = max(0, plot_start)

    if plot_stride is None:
        plot_stride = prediction_length

    # if plot_start > context_length:
    #     logging.warning(
    #         f"To see the plot lines of the prediction and true values, "
    #         f"choose the plot_start less than {context_length}."
    #     )

    if plot_type == "plotly":
        fig = make_subplots(specs=[[{"secondary_y": True}]])
    else:
        fig = plt.figure(figsize=(fig_size[0], fig_size[1]))
    for c in forecast_columns:
        x = test_data_updated[timestamp_column].iloc[plot_start:plot_end]
        y = test_data_updated[c].iloc[plot_start:plot_end]
        plot_line(x=x, y=y, plot_type=plot_type, plot_id=c, fig=fig)

    # plot prediction data

    predictions_start = max(0, plot_start - context_length)
    # predictions_end = max(plot_end - context_length, 0)

    # index into the predictions so that the end of the prediction coincides with the end of the ground truth
    #
    predictions_end = plot_range[-1] - prediction_length - context_length + 1  #  - context_length - prediction_length

    predictions_start = plot_range[0] - context_length

    plot_index = range(predictions_end, predictions_start - 1, -plot_stride)

    if num_predictions and num_predictions < len(plot_index):
        plot_index = plot_index[:num_predictions]

    for c in forecast_columns:
        forecast_name = f"{c}_prediction"
        if plot_type == "plotly":
            for i in plot_index:
                start = forecast_data.iloc[i][timestamp_column]
                timestamps = pd.date_range(start, freq=periodicity, periods=prediction_length + 1)
                timestamp = timestamps[1:]
                forecast_val = forecast_data.iloc[i][forecast_name]
                plot_line(
                    x=timestamp,
                    y=forecast_val,
                    plot_type=plot_type,
                    plot_id=forecast_name,
                    fig=fig,
                )
        else:
            test_forecast = explode_and_add_timestamps(
                forecast_data[[timestamp_column, forecast_name]],
                forecast_name,
                timestamp_column,
                periodicity,
            )
            plot_line(
                data=test_forecast,
                x="new_timestamp",
                y=forecast_name,
                plot_type=plot_type,
                plot_id=forecast_name,
                estimator="median",
                errorbar=("ci", 95),
            )

    # final plot formatting
    if plot_type == "plotly":
        fig["layout"].update(
            height=fig_size[1],
            width=fig_size[0],
            title=title,
            xaxis={"tickangle": -45},
        )
        if return_image:
            return Image(fig.to_image(format="png"))
        else:
            return fig
    else:
        plt.title(title)
        plt.xticks(rotation=45)
        plt.close()
        return fig


def plot_predictions(
    input_df: Optional[pd.DataFrame] = None,
    predictions_df: Optional[pd.DataFrame] = None,
    exploded_predictions_df: Optional[pd.DataFrame] = None,
    dset: Optional[Dataset] = None,
    model: Optional[PreTrainedModel] = None,
    freq: Optional[str] = None,
    timestamp_column: Optional[str] = None,
    id_columns: Optional[List[str]] = None,
    plot_context: Optional[int] = None,
    plot_dir: str = None,
    num_plots: int = 10,
    plot_prefix: str = "valid",
    channel: Union[int, str] = None,
    indices: List[int] = None,
):
    """Utility for plotting forecasts along with context and test data.

    User should pass either:

        - input_df and predictions_df: context will be extracted from input_df, and predictions will be extracted from
            predictions_df. Predictions_df is expected to have rows containing lists of predictions.
        - input_df and exploded_predictions_df: context will be extracted from input_df, and predictions from
            exploded_predictions_df will be plotted
        - dset and model: model will be used to produce predictions from records selected from dset

    If exploded_predictions_df is passed, indices and num_plots are ignored, the assumption is that there are only one
        set of predictions passed for plotting.

    Args:
        input_df (Optional[pd.DataFrame], optional): The input dataframe from which the predictions are generated,
            containing timestamp and target columns. Defaults to None.
        predictions_df (Optional[pd.DataFrame], optional): The predictions dataframe, where each row contains starting
            timestamp and a list of predictions for each target column. Defaults to None.
        exploded_predictions_df (Optional[pd.DataFrame], optional): The predictions dataframe, containing timestamp
            and predicted target columns. Defaults to None.
        dset (Optional[Dataset], optional): Torch dataset containing the context data to use as input for the model.
            Defaults to None.
        model (Optional[PreTrainedModel], optional): The pre-trained time series model. Defaults to None.
        freq (Optional[str], optional): Frequency of the time series data, using Pandas string abbreviations
            (https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases). Defaults to None.
        timestamp_column (Optional[str], optional): Name of timestamp column in the dataframe. Defaults to None.
        id_columns (Optional[List[str]], optional):  (For future use) List of id columns in the dataframe. Defaults to
            None.
        plot_context (Optional[int], optional): Integer representing the number of time points of historical data to
            plot. Defaults to None.
        plot_dir (str, optional): Directory where plots are saved. Defaults to None.
        num_plots (int, optional): Number of subplots to plot in the figure. Defaults to 10.
        plot_prefix (str, optional): Prefix to put on the plot file names. Defaults to "valid".
        channel (Union[int, str], optional): Channel, i.e., target column or its index, to plot. Defaults to None.
        indices (List[int], optional): List of indices to plot. If None, random examples will be chosen. Defaults to
            None.
    """
    if indices is not None:
        num_plots = len(indices)

    # possible operations:
    if input_df is not None and exploded_predictions_df is not None:
        # 1) This is a zero-shot prediction, so no test data. We have context data for the channel (target column).
        # We expect the context and predictions to contain the channel
        pchannel = f"{channel}_prediction"
        if pchannel not in exploded_predictions_df.columns:
            raise ValueError(f"Predictions dataframe does not contain target column '{pchannel}'.")
        if channel not in input_df.columns:
            raise ValueError(f"Context dataframe does not contain target column '{channel}'.")

        num_plots = 1
        prediction_length = len(exploded_predictions_df)
        plot_context = len(input_df)
        using_pipeline = True
        plot_test_data = False
        indices = [-1]  # indices not used in exploded case
    elif input_df is not None and predictions_df is not None:
        # 2) input_df and predictions plus column information is provided
        pchannel = f"{channel}_prediction" if f"{channel}_prediction" in predictions_df else channel

        if indices is None:
            l = len(predictions_df)
            num_plots = min(num_plots, l)
            indices = np.random.choice(l, size=num_plots, replace=False)
        predictions_subset = [predictions_df.iloc[i] for i in indices]

        gt_df = input_df.copy()
        gt_df = gt_df.set_index(timestamp_column)  # add id column logic here

        prediction_length = len(predictions_subset[0][pchannel])
        using_pipeline = True
        plot_test_data = True
    elif model is not None and dset is not None:
        # 3) model and dataset are provided
        device = model.device

        with torch.no_grad():
            if indices is None:
                num_plots = min(num_plots, len(dset))
                indices = np.random.choice(len(dset), size=num_plots, replace=False)

            signature = inspect.signature(model.forward)
            signature_keys = list(signature.parameters.keys())
            dset_keys = dset[0].keys()
            random_samples = {}
            for k in dset_keys:
                if k in signature_keys:
                    random_samples[k] = torch.stack([dset[i][k] for i in indices]).to(device=device)
            output = model(**random_samples)
            predictions_subset = output.prediction_outputs[:, :, channel].squeeze().cpu().numpy()
            prediction_length = predictions_subset.shape[1]
        using_pipeline = False
        plot_test_data = True
    else:
        raise RuntimeError(
            "You must provide either input_df and predictions_df, or dset and model, or input_df and exploded_predictions_df."
        )

    if plot_context is None:
        plot_context = 2 * prediction_length

    # Set a more beautiful style
    plt.style.use("seaborn-v0_8-whitegrid")

    # Adjust figure size and subplot spacing
    assert num_plots >= 1
    fig, axs = plt.subplots(num_plots, 1, figsize=(10, 2 * num_plots))
    if num_plots == 1:
        axs = [axs]

    for i, index in enumerate(indices):
        if using_pipeline and plot_test_data:
            ts_y_hat = create_timestamps(predictions_subset[i][timestamp_column], freq=freq, periods=prediction_length)
            y_hat = (
                predictions_subset[i][f"{channel}_prediction"]
                if f"{channel}_prediction" in predictions_subset[i]
                else predictions_subset[i][channel]
            )

            # get ground truth
            loc = gt_df.index.get_loc(predictions_subset[i][timestamp_column])
            ts_index = gt_df.index[loc - plot_context + 1 : loc + 1 + prediction_length]
            y = gt_df.loc[ts_index][channel]
            ts_y = y.index
            y = y.values
            # border = ts_y[-prediction_length]
            border = predictions_subset[i][timestamp_column]
            plot_title = f"Example {indices[i]}"

        elif using_pipeline:
            ts_y_hat = create_timestamps(
                exploded_predictions_df[timestamp_column].iloc[0], freq=freq, periods=prediction_length
            )
            y_hat = exploded_predictions_df[f"{channel}_prediction"]

            # get context
            # ts_y = create_timestamps(context_df[timestamp_column].iloc[0], freq=freq, periods=len(context_df))
            ts_y = input_df[timestamp_column].values
            y = input_df[channel].values
            border = None
            plot_title = f"Forecast for {channel}"

        else:
            batch = dset[index]
            feasible_plot_context = min(plot_context, batch["past_values"].shape[0])
            ts_y_hat = np.arange(feasible_plot_context, feasible_plot_context + prediction_length)
            y_hat = predictions_subset[i]

            ts_y = np.arange(feasible_plot_context + prediction_length)
            y = batch["future_values"][:, channel].squeeze().numpy()
            x = batch["past_values"][-feasible_plot_context:, channel].squeeze().numpy()
            y = np.concatenate((x, y), axis=0)
            border = feasible_plot_context
            plot_title = f"Example {indices[i]}"

        # Plot predicted values with a dashed line
        axs[i].plot(ts_y_hat, y_hat, label="Predicted", linestyle="--", color="orange", linewidth=2)

        # Plot true values with a solid line
        axs[i].plot(ts_y, y, label="True", linestyle="-", color="blue", linewidth=2)

        # Plot horizon border
        if border is not None:
            axs[i].axvline(x=border, color="r", linestyle="-")

        axs[i].set_title(plot_title)
        axs[i].legend()

    # Adjust overall layout
    plt.tight_layout()

    # Save the plot
    if plot_dir is not None:
        plot_filename = f"{plot_prefix}_ch_{str(channel)}.pdf"
        os.makedirs(plot_dir, exist_ok=True)
        plt.savefig(os.path.join(plot_dir, plot_filename))



# Copyright contributors to the TSFM project
#

"""Tests for HF config registration"""

import pytest

from tsfm_public.toolkit.hf_util import register_config


def test_register_config():
    with pytest.raises(Exception) as ex:
        register_config("patchtst", "PatchTSTConfig", "UnkownRandomModule")
    assert "Could not load PatchTSTConfig from" in str(ex)

    with pytest.raises(Exception) as ex:
        register_config("patchtst", "PatchTSTConfig2", "transformers")

    assert "Could not find config for PatchTSTConfig2" in str(ex)



# Copyright contributors to the TSFM project
#
"""Helper class for anomaly detection support"""

from collections import OrderedDict
from typing import List, Optional

import numpy as np
import torch
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler as MinMaxScaler_
from sklearn.preprocessing import StandardScaler as StandardScaler_
from torch import nn as nn
from transformers.utils.generic import ModelOutput

from tsfm_public.models.tspulse.modeling_tspulse import TSPulseForReconstruction
from tsfm_public.toolkit.ad_helpers import (
    AnomalyScoreMethods,
    ScoreListType,
    TSADHelperUtility,
)

from .helpers import patchwise_stitched_reconstruction


class TSPulseADUtility(TSADHelperUtility):
    """Implements TSAD Helper Utility for TSPulse model"""

    def __init__(
        self,
        model: TSPulseForReconstruction,
        mode: List[str],
        aggregation_length: int,
        score_exponent: float = 1.0,
        least_significant_scale: float = 1e-2,
        least_significant_score: float = 0.2,
        **kwargs,
    ):
        """Initializer

        Args:
            model (TSPulseForReconstruction): model instance.
            mode (list): mode string specifies scoring logic.
            aggregation_length (int): parameter for imputation based scoring.
            score_exponent (float, optional): parameter to sharpen the anomaly score. Defaults to 1.
            least_significant_scale (float, optional): allowed model deviation from the data in the scale of data variance. Defaults to 1e-2.
            least_significant_score (float, optional): minimum anomaly score for significant detection. Defaults to 0.2.

        Raises:
            ValueError: unsupported scoring mode
        """
        if mode is None:
            mode = [
                AnomalyScoreMethods.PREDICTIVE.value,
                AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value,
                AnomalyScoreMethods.TIME_RECONSTRUCTION.value,
            ]

        super(TSPulseADUtility, self).__init__()
        if not self.is_valid_mode(mode):
            raise ValueError(f"Error: unsupported inference method {mode}!")
        self._model = model
        self._mode = mode
        self._aggr_win_size = aggregation_length
        self._score_exponent = score_exponent
        self._least_significant_scale = least_significant_scale
        self._least_significant_score = least_significant_score

    def is_valid_mode(self, mode_str: List[str]) -> bool:
        """Validates compatibility of the specified mode string."""
        supported_modes = [
            AnomalyScoreMethods.PREDICTIVE.value,
            AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value,
            AnomalyScoreMethods.TIME_RECONSTRUCTION.value,
        ]

        valid_mode = False
        for mode_type in supported_modes:
            if mode_type in mode_str:
                valid_mode = True
        return valid_mode

    def preprocess(self, x: DataFrame, **kwargs) -> DataFrame:
        """Performs standard normalization on the target columns before scoring.

        Args:
            x (DataFrame): input for scoring

        Returns:
            DataFrame: processed dataframe
        """
        x = super().preprocess(x, **kwargs)
        x_ = x.copy()
        target_columns = kwargs.get("target_columns", [])
        if len(target_columns) > 0:
            x_[target_columns] = StandardScaler_().fit_transform(x[target_columns].values)
        return x_

    def compute_score(
        self,
        payload: dict,
        expand_score: bool = False,
        **kwargs,
    ) -> ModelOutput:
        """Produces required model output for anomaly scoring

        Args:
            payload (dict): data batch.
            expand_score (bool): compute score for each stream for multivariate data. Defaults to False.

        Returns:
            ModelOutput: model output
        """
        mode = kwargs.get("mode", self._mode)
        use_forecast = AnomalyScoreMethods.PREDICTIVE.value in mode
        use_fft = AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value in mode
        use_ts = AnomalyScoreMethods.TIME_RECONSTRUCTION.value in mode
        aggr_win_size = self._aggr_win_size
        anomaly_criterion = nn.MSELoss(reduction="none")

        reconstruct_start = self._model.config.context_length - aggr_win_size
        reconstruct_end = self._model.config.context_length

        batch_x = payload["past_values"]

        # Get TSPulse zeroshot output with stitched masked reconstruction
        keys_to_stitch = ["reconstruction_outputs", "reconstructed_ts_from_fft"]

        model_forward_output = {}
        if use_forecast:
            # model_forward_output = self._model(**payload)
            model_forward_output = self._model(batch_x)

        stitched_dict = {}
        if use_ts or use_fft:
            stitched_dict = patchwise_stitched_reconstruction(
                model=self._model,
                past_values=batch_x,
                patch_size=self._model.config.patch_length,
                keys_to_stitch=keys_to_stitch,
                keys_to_aggregate=[],
                reconstruct_start=reconstruct_start,
                reconstruct_end=reconstruct_end,
                debug=False,
            )
            if isinstance(stitched_dict, tuple):
                stitched_dict = stitched_dict[0]

        # Get desired output from TSPulse outputs
        # output shape: [batch_size, window_size, n_channels]
        scores = OrderedDict()

        reduction_axis = [1] if expand_score else [1, 2]
        if use_ts:
            # time reconstruction
            output = stitched_dict["reconstruction_outputs"]
            pointwise_score = anomaly_criterion(
                batch_x[:, reconstruct_start:reconstruct_end, :],
                output[:, reconstruct_start:reconstruct_end, :],
            )
            scores[AnomalyScoreMethods.TIME_RECONSTRUCTION.value] = torch.mean(pointwise_score, dim=reduction_axis)

        if use_fft:
            # time reconstruction from fft
            output = stitched_dict["reconstructed_ts_from_fft"]
            pointwise_score = anomaly_criterion(
                batch_x[:, reconstruct_start:reconstruct_end, :],
                output[:, reconstruct_start:reconstruct_end, :],
            )
            scores[AnomalyScoreMethods.FREQUENCY_RECONSTRUCTION.value] = torch.mean(
                pointwise_score, dim=reduction_axis
            )

        if use_forecast:
            # forecast output
            batch_future_values = payload["future_values"]
            output = model_forward_output["forecast_output"]
            pointwise_score = anomaly_criterion(batch_future_values[:, 0, :], output[:, 0, :]).unsqueeze(1)
            scores[AnomalyScoreMethods.PREDICTIVE.value] = torch.mean(pointwise_score, dim=reduction_axis)

        return ModelOutput(scores)

    def adjust_boundary(
        self,
        key: str,
        x: ScoreListType,
        reference: Optional[np.ndarray] = None,
        **kwargs,
    ) -> np.ndarray:
        """Combines model outputs with boundary adjustment.

        Args:
            key (str): key associated with model output.
            x (ScoreListType): model outputs across all batches combined.
            reference (np.ndarray, optional): reference data for score scale adjustment. Defaults to None.

        Returns:
            np.ndarray: combined score
        """
        context_length = self._model.config.context_length
        aggr_win_size = self._aggr_win_size
        score_exponent = self._score_exponent
        if isinstance(x, (list, tuple)):
            if (len(x) > 0) and isinstance(x[0], torch.Tensor):
                x = torch.cat(x, axis=0).detach().cpu().numpy()
            else:
                x = np.concatenate(x, axis=0).astype(float)
        if key == AnomalyScoreMethods.PREDICTIVE.value:
            start_pad_len = context_length
            end_pad_len = 0
        else:
            start_pad_len = context_length - aggr_win_size // 2
            end_pad_len = aggr_win_size // 2

        score = np.array([x[0]] * start_pad_len + list(x) + [x[-1]] * end_pad_len)
        if score.ndim == 1:
            score = score.reshape(-1, 1)

        min_score = 0.0
        if reference is not None:
            reference_data = np.asarray(reference)
            min_score = (
                self._least_significant_scale * np.nanstd(np.diff(reference_data, axis=0), axis=0, keepdims=True) ** 2
            )

            if min_score.shape[-1] != score.shape[-1]:
                min_score = np.nanmax(min_score, axis=-1)
            if key == AnomalyScoreMethods.PREDICTIVE.value:
                min_score = min_score * np.sqrt(2)
            else:
                min_score = min_score * (1 + self._model.config.patch_length)

        score_ = score.copy()
        score_[np.where(score > min_score)] *= 1 / self._least_significant_score
        scale = 1 if np.any(score > min_score) else self._least_significant_score
        score = MinMaxScaler_().fit_transform(score_**score_exponent) * scale
        return score



import os


# this needs to come before any tsfm imports
os.environ["TSFM_MODEL_DIR"] = "granite-tsfm/services/inference/mytest-tsfm"

import json  # noqa: E402
import shutil  # noqa: E402
import tarfile  # noqa: E402
from datetime import datetime  # noqa: E402

import numpy as np  # noqa: E402
import pandas as pd  # noqa: E402
from inference import input_fn, model_fn, output_fn, predict_fn  # noqa: E402
from tsfminference.inference_payloads import PredictOutput  # noqa: E402


def payload(json_file_name: str = None):
    if json_file_name:
        return open(json_file_name).read()

    # Generate 10 sequential dates starting from today
    tslength = 1024
    start_date = datetime(2020, 1, 1)
    date_range = [d.isoformat() for d in pd.date_range(start=start_date, periods=tslength, freq="h")]
    ids = ["A" for _ in range(tslength)]
    values = np.random.rand(tslength)

    # Create the DataFrame
    df = pd.DataFrame({"date": date_range, "ID": ids, "value": values})

    return json.dumps(
        {
            "inference_type": "forecasting",  # we currently support only 'forecasting'
            "model_id": "ttm-1024-96-r1",
            "parameters": {},
            "schema": {
                "timestamp_column": "date",
                "id_columns": ["ID"],  # multiple columns are supported
                "target_columns": ["value"],  # what we're generating a forecast for
            },
            "data": df.to_dict(orient="list"),
            "future_data": {},  # used for things like exogenous data
        }
    )


def fetch_model(model_data):
    """Untar the model.tar.gz object either from local file system
    or a S3 location

    Args:
        model_data (str): either a path to local file system starts with
        file:/// that points to the `model.tar.gz` file or an S3 link
        starts with s3:// that points to the `model.tar.gz` file

    Returns:
        model_dir (str): the directory that contains the uncompress model
        checkpoint files
    """

    # no user serviceable parts here!
    # this just emulates what sagemaker would do
    # on the cloud

    model_dir = "/tmp/model"
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    # _check_model(model_data)
    shutil.copy2(
        model_data,
        model_dir,
    )

    # untar the model
    tar = tarfile.open(os.path.join(model_dir, "model.tar.gz"))
    tar.extractall(model_dir)
    tar.close()

    return model_dir


def test(model_data, json_file_name: str = None):
    # decompress the model.tar.gz file
    model_dir = fetch_model(model_data)

    # load the model
    net = model_fn(model_dir)
    jstr = payload(json_file_name=json_file_name)
    # encode numpy array to binary stream
    content_type = "application/json"
    input_object = input_fn(jstr, content_type)
    predictions = predict_fn(input_object, net)

    if not isinstance(predictions, PredictOutput):
        print(f"non-normal return {predictions}")
        sys.exit(1)

    res = output_fn(predictions, content_type)
    res = json.loads(res)
    assert res["output_data_points"] == 96


if __name__ == "__main__":
    model_data = "../model.tar.gz"

    import sys

    json_file_name = None
    if len(sys.argv) > 1:
        json_file_name = sys.argv[1]

    test(model_data, json_file_name)



# Copyright contributors to the TSFM project
#

"""Tests get_model"""

import numpy as np
import pytest

from tsfm_public.toolkit.get_model import get_model


def test_granite_r1_models():
    # Check basic cases
    mp = "ibm-granite/granite-timeseries-ttm-r1"
    for cl in [512, 1024]:
        for fl in [96]:
            model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
            assert int(model_key.split("-")[1]) == fl
            assert int(model_key.split("-")[0]) == cl

    mp = "ibm-granite/granite-timeseries-ttm-r1"
    cl = 1024
    fl = 56
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        head_dropout=0.3,
        return_model_key=True,
    )

    mp = "ibm/TTM"
    cl = 512
    fl = 90
    model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
    assert int(model_key.split("-")[0]) == cl
    assert int(model_key.split("-")[1]) == 96

    mp = "ibm/TTM"
    cl = 300
    fl = 45
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_longer_context=False,
        return_model_key=True,
        force_return="zeropad",
    )
    assert int(model_key.split("-")[0]) == 512
    assert int(model_key.split("-")[1]) == 96


def test_research_r2_models():
    mp = "ibm-research/ttm-research-r2"
    for cl in [512, 1024, 1536]:
        for fl in [96, 192, 336, 720]:
            model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
            assert int(model_key.split("-")[1]) == fl
            assert int(model_key.split("-")[0]) == cl


def test_granite_r2_basic_models():
    mp = "ibm-granite/granite-timeseries-ttm-r2"
    # 12 high freq models
    for cl in [512, 1024, 1536]:
        for fl in [96, 192, 336, 720]:
            model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
            assert int(model_key.split("-")[1]) == fl
            assert int(model_key.split("-")[0]) == cl

    # ---- Low freq models [52, 90, 180, 360, 512] ----
    # 52 - l2 loss
    cl = 52
    fl = 16
    model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl

    # 52 - l1 loss
    cl = 52
    fl = 16
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # 90 - l2 loss
    cl = 90
    fl = 30
    model_key = get_model(model_path=mp, context_length=cl, prediction_length=fl, return_model_key=True)
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl

    # 90 - l1 loss
    cl = 90
    fl = 30
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # 180 - l1 loss
    cl = 180
    fl = 60
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # 360 - l1 loss
    cl = 360
    fl = 60
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # 512/48 - l2 loss
    cl = 512
    fl = 48
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=False, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl

    # 512/48 - l1 loss
    cl = 512
    fl = 48
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_l1_loss=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # 512/96 - l2 loss
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=False,
        freq_prefix_tuning=True,
        freq="d",
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl

    # 512/96 - l1 loss
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=True,
        freq_prefix_tuning=True,
        freq="10min",
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key


def test_granite_r2_other_models():
    mp = "ibm-granite/granite-timeseries-ttm-r2"
    cl = 512
    fl = 10
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        dropout=0.4,
        decoder_num_layers=1,
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == 48
    assert int(model_key.split("-")[0]) == cl

    mp = "ibm-granite/granite-timeseries-ttm-r2"
    cl = 1536
    fl = 200
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        decoder_adaptive_patching_levels=2,
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == 336
    assert int(model_key.split("-")[0]) == cl


def test_prefer_longer_context():
    mp = "ibm-granite/granite-timeseries-ttm-r2"
    # Case 1
    cl = 100
    fl = 20
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_longer_context=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == 30
    assert int(model_key.split("-")[0]) == 90

    # Case 2
    cl = 100
    fl = 20
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_longer_context=False, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == 30
    assert int(model_key.split("-")[0]) == 90

    # Case 3
    cl = 100
    fl = 10
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_longer_context=False, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == 16
    assert int(model_key.split("-")[0]) == 52

    # Case 4
    cl = 100
    fl = 10
    model_key = get_model(
        model_path=mp, context_length=cl, prediction_length=fl, prefer_longer_context=True, return_model_key=True
    )
    assert int(model_key.split("-")[1]) == 30
    assert int(model_key.split("-")[0]) == 90


def test_freq_tuning():
    mp = "ibm-granite/granite-timeseries-ttm-r2"
    # # ----------- test freq_prefix_tuning -----------
    # Case 1
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=False,
        freq_prefix_tuning=False,
        freq="h",
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl

    # Case 2
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=True,
        freq_prefix_tuning=False,
        freq="H",
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key

    # Case 3
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=False,
        freq_prefix_tuning=True,
        freq="3min",
        return_model_key=True,
        force_return="random_init_large",
    )
    assert "large" in model_key

    # Case 4
    cl = 512
    fl = 96
    model_key = get_model(
        model_path=mp,
        context_length=cl,
        prediction_length=fl,
        prefer_l1_loss=True,
        freq_prefix_tuning=True,
        freq="H",
        return_model_key=True,
    )
    assert int(model_key.split("-")[1]) == fl
    assert int(model_key.split("-")[0]) == cl
    assert "l1" in model_key


def test_random_models():
    mp = "ibm-granite/granite-timeseries-ttm-r2"

    for cl in np.linspace(2, 20_000, 10):
        for fl in [cl // 8, cl // 4, cl // 2]:
            fl = int(fl)
            cl = int(cl)
            if fl < 1:
                continue
            if cl < 52:
                force_return = "zeropad"
            if fl > 720:
                force_return = "rolling"
            model_key = get_model(
                model_path=mp,
                context_length=cl,
                prediction_length=fl,
                return_model_key=True,
                force_return=force_return,
            )
            assert model_key is not None


# Define test cases with multiple combinations
# (cl, fl, res, ft, l1, longer, force_return)
test_cases_granite_r2 = [
    # Basic exact matches
    (512, 96, None, False, False, True, None, "512-96-r2"),
    (1024, 192, None, False, False, True, None, "1024-192-r2"),
    (1536, 720, None, False, False, True, None, "1536-720-r2"),
    # Forecast length filtering
    (512, 720, None, False, False, True, "rolling", "512-720-r2"),
    (1024, 1000, None, False, False, True, "rolling", "1024-720-r2"),
    (1024, 1000, None, False, False, True, "rolling", "1024-720-r2"),
    (512, 192, None, False, False, True, "rolling", "512-192-r2"),
    (180, 60, None, False, False, True, "rolling", "180-60-ft-l1-r2.1"),
    (200, 50, None, False, False, True, "rolling", "180-60-ft-l1-r2.1"),
    # Context length filtering
    (2000, 336, None, False, False, True, "zeropad", "1536-336-r2"),
    (1200, 96, None, False, False, True, "zeropad", "1024-96-r2"),
    (600, 336, None, False, False, True, "zeropad", "512-336-r2"),
    # Context length ordering (Prefer Longer Context)
    (1200, 96, None, False, False, True, None, "1024-96-r2"),
    (1200, 96, None, False, False, False, None, "512-96-r2"),
    (80, 20, "10min", False, False, True, "rolling", "52-16-ft-r2.1"),
    (80, 200, "10min", False, False, True, "random_init_small", "TTM(small)"),
    (80, 40, "10min", False, False, True, "rolling", "52-16-ft-r2.1"),
    (200, 80, "10min", False, False, False, "rolling", "180-60-ft-l1-r2.1"),
    (200, 80, "10min", False, False, True, "rolling", "180-60-ft-l1-r2.1"),
    (400, 80, "10min", False, False, False, "rolling", "180-60-ft-l1-r2.1"),
    (400, 80, "10min", False, False, True, "rolling", "360-60-ft-l1-r2.1"),
    # FT and L1 Preferences (Only Apply for CL ≤ 512)
    (512, 96, None, False, True, True, None, "512-96-ft-l1-r2.1"),
    (1024, 96, None, False, True, True, None, "1024-96-r2"),  # L1 ignored
    (512, 96, None, True, False, True, None, "512-96-ft-r2.1"),
    (1024, 96, None, True, False, True, None, "1024-96-r2"),  # FT ignored
    (512, 96, None, True, True, True, None, "512-96-ft-l1-r2.1"),
    # Resolution-based filtering
    (512, 96, "d", False, False, True, None, "512-96-ft-r2.1"),
    (512, 96, "d", False, True, True, None, "512-96-ft-l1-r2.1"),
    (300, 20, "W", True, True, True, "zeropad", "180-60-ft-l1-r2.1"),
    (60, 12, "W", True, True, True, "zeropad", "52-16-ft-l1-r2.1"),
    (60, 12, "W", True, False, True, "zerop", "52-16-ft-r2.1"),
    (36, 12, "W", True, True, False, "zeropad", "52-16-ft-l1-r2.1"),
    (36, 12, "W", False, False, False, "zeropad", "52-16-ft-r2.1"),
    (36, 12, "M", False, False, True, "random_init_small", "TTM(small)"),
    (512, 96, "oov", False, False, True, None, "512-96-r2"),
    (512, 96, "5min", False, False, True, None, "512-96-r2"),
    (512, 96, "random", False, False, True, None, "512-96-r2"),  # Invalid freq, but ft=False
    (512, 96, "random", True, False, True, "random_init_medium", "TTM(medium)"),  # Invalid freq, but ft=True
    (20, 6, "W", True, True, False, "zeropad", "52-16-ft-l1-r2.1"),
    (200, 24, "W", True, False, True, "zeropad", "180-60-ft-l1-r2.1"),
    (200, 24, "W", True, False, False, "zeropad", "90-30-ft-r2.1"),
    (200, 24, "H", True, False, False, "zeropad", "90-30-ft-r2.1"),
    (20, 6, "A", True, True, True, "random_init_small", "TTM(small)"),
    # Edge Cases (No Match Scenarios + some match scenarios)
    (10, 5, None, False, False, True, "random_init_small", "TTM(small)"),
    (10, 5, None, True, False, True, "random_init_small", "TTM(small)"),
    (10, 5, None, False, False, True, "zeropad", "1536-96-r2"),
    (10, 5, None, False, False, False, "zeropad", "52-16-ft-r2.1"),
    (10, 5, None, True, False, False, "zeropad", "52-16-ft-r2.1"),
    (100, 50, None, True, False, True, "random_init_small", "TTM(small)"),
    (100, 50, None, True, False, True, "random_init_medium", "TTM(medium)"),
    (100, 50, None, True, True, True, "rolling", "90-30-ft-l1-r2.1"),
    (600, 50, None, False, False, True, None, "512-96-r2"),
    (600, 40, None, False, False, True, None, "512-48-ft-r2.1"),
    (1536, 1000, None, False, False, True, "random_init_large", "TTM(large)"),
    (1536, 336, "W", False, False, True, "random_init_small", "TTM(small)"),
    (1536, 96, "d", True, True, True, None, "512-96-ft-l1-r2.1"),
    (13, 3, "d", True, True, True, "random_init_small", "TTM(small)"),
    # Complex Cases with Multiple Constraints
    (512, 96, "oov", True, True, True, None, "512-96-ft-l1-r2.1"),
    (512, 192, "d", True, False, False, "rolling", "512-96-ft-r2.1"),
    (512, 192, "d", True, False, False, "random_init_small", "TTM(small)"),
    (1024, 96, None, True, True, False, None, "512-96-ft-l1-r2.1"),
    (1536, 336, "d", False, False, True, "rolling", "512-96-ft-r2.1"),
    (1536, 336, "d", False, True, True, "rolling", "512-96-ft-l1-r2.1"),
    # Sorting correctness
    (1200, 500, None, False, False, True, None, "1024-720-r2"),
    (1200, 500, None, False, False, False, None, "512-720-r2"),
    (1200, 500, None, False, True, False, "rolling", "512-96-ft-l1-r2.1"),
    (1200, 500, None, False, True, False, "random_init_small", "TTM(small)"),
]


@pytest.mark.parametrize("cl, fl, res, ft, l1, longer, force_return, expected", test_cases_granite_r2)
def test_all_cases_granite_r2(cl, fl, res, ft, l1, longer, force_return, expected):
    model_key = get_model(
        "ibm-granite/granite-timeseries-ttm-r2",
        context_length=cl,
        prediction_length=fl,
        freq=res,
        freq_prefix_tuning=ft,
        prefer_l1_loss=l1,
        prefer_longer_context=longer,
        force_return=force_return,
        return_model_key=True,
    )
    assert model_key == expected


# (cl, fl, res, ft, l1, longer, force_return)
test_cases_granite_r1 = [
    # Basic exact matches
    (512, 96, None, False, False, True, None, "512-96-r1"),
    (1024, 96, None, False, False, True, None, "1024-96-r1"),
    # filter prediction length
    (512, 50, None, False, False, False, None, "512-96-r1"),
    # invalid context length
    (50, 50, None, False, False, False, "zeropad", "512-96-r1"),
    (50, 50, None, False, False, False, "random_init_small", "TTM(small)"),
    # prefer longer context
    (1200, 50, None, False, False, True, None, "1024-96-r1"),
    (1200, 50, None, False, False, False, None, "512-96-r1"),
]


@pytest.mark.parametrize("cl, fl, res, ft, l1, longer, force_return, expected", test_cases_granite_r1)
def test_all_cases_granite_r1(cl, fl, res, ft, l1, longer, force_return, expected):
    model_key = get_model(
        "ibm-granite/granite-timeseries-ttm-r1",
        context_length=cl,
        prediction_length=fl,
        freq=res,
        freq_prefix_tuning=ft,
        prefer_l1_loss=l1,
        prefer_longer_context=longer,
        force_return=force_return,
        return_model_key=True,
    )
    assert model_key == expected


test_cases_research_r2 = [
    # Basic exact matches
    (512, 96, None, False, False, True, None, "512-96-ft-r2"),
    (1024, 96, None, False, False, True, None, "1024-96-ft-r2"),
    # filter prediction length
    (512, 50, None, False, False, False, None, "512-96-ft-r2"),
    # invalid context length
    (50, 50, None, False, False, False, "zeropad", "512-96-ft-r2"),
    (50, 50, None, False, False, False, "random_init_small", "TTM(small)"),
    # prefer longer context
    (1200, 50, None, False, False, True, None, "1024-96-ft-r2"),
    (1200, 50, None, False, False, False, None, "512-96-ft-r2"),
]


@pytest.mark.parametrize("cl, fl, res, ft, l1, longer, force_return, expected", test_cases_research_r2)
def test_all_cases_research_r2(cl, fl, res, ft, l1, longer, force_return, expected):
    model_key = get_model(
        "ibm-research/ttm-research-r2",
        context_length=cl,
        prediction_length=fl,
        freq=res,
        freq_prefix_tuning=ft,
        prefer_l1_loss=l1,
        prefer_longer_context=longer,
        force_return=force_return,
        return_model_key=True,
    )
    assert model_key == expected
